"""Support for the MSBG hearing loss model."""
import json
import logging
import math
import os
from typing import Dict, Optional, Tuple, Union

import numpy as np
import scipy
import scipy.signal
import soundfile
from soundfile import SoundFile

# measure rms parameters
WIN_SECS = 0.01

# read & write signal parameters
MSBG_FS = 44100
TEST_NBITS = 16

# fmt: off
HZ = np.array(
    [
        0.0, 20.0, 25.0, 31.5, 40.0, 50.0, 63.0, 80.0, 100.0, 125.0, 160.0, 200.0,
        250.0, 315.0, 400.0, 500.0, 630.0, 750.0, 800.0, 1000.0, 1250.0, 1500.0,
        1600.0, 2000.0, 2500.0, 3000.0, 3150.0, 4000.0, 5000.0, 6000.0, 6300.0,
        8000.0, 9000.0, 10000.0, 11200.0, 12500.0, 14000.0, 15000.0, 16000.0,
        20000.0, 48000,
    ]
)

MIDEAR = np.array(
    [
        50.0, 39.6, 32.0, 25.85, 21.4, 18.5, 15.9, 14.1, 12.4, 11.0, 9.6, 8.3, 7.4,
        6.2, 4.8, 3.8, 3.3, 2.9, 2.6, 2.6, 4.5, 5.4, 6.1, 8.5, 10.4, 7.3, 7.0, 6.6,
        7.0, 9.2, 10.2, 12.2, 10.8, 10.1, 12.7, 15.0, 18.2, 23.8, 32.3, 50.0, 50.0,
    ]
)

# Free field (frontal)FF_ED correction for threshold (was ISO std Table 1 - 4.2 dB)
# i.e. relative to 0.0 dB at 1000 Hz, Shaw 1974

FF_ED = np.array(
    [
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.3, 0.5, 0.9, 1.4,
        1.6, 1.7, 2.5, 2.7, 2.6, 2.6, 3.2, 5.2, 6.6, 12.0, 16.8, 15.3, 15.2, 14.2,
        10.7, 7.1, 6.4, 1.8, -0.9, -1.6, 1.9, 4.9, 2.0, -2.0, 2.5, 2.5, 2.5,
    ]
)

# DIFFUSE field ( relative to 0.0 dB at 1000Hz)
# from 2008 file [corrections08.m] used in Samsung collaboration

DF_ED = np.array(
    [
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.3, 0.4, 0.5, 1.0,
        1.6, 1.7, 2.2, 2.7, 2.9, 3.8, 5.3, 6.8, 7.2, 10.2, 14.9, 14.5, 14.4,
        12.7, 10.8, 8.9, 8.7, 8.5, 6.2, 5.0, 4.5, 4.0, 3.3, 2.6, 2.0, 2.0, 2.0,
    ]
)

# ITU Rec P 58 08/96 Head and Torso Simulator transfer fns. from Peter Hugher BTRL,
# 4-June-2001. Negative of values in Table 14a of ITU P58 (05/2013), accessible at
# http://www.itu.int/rec/T-REC-P.58-201305-I/en
# Freely available. Converts from ear reference point (ERP) to eardrum reference
# point (DRP). EXCEPT extra 2 points added for 20k & 48k by MAS, MAr 2012

ITU_HZ = np.array(
    [
        0, 100, 125, 160, 200, 250, 315, 400, 500, 630, 800, 1000, 1250, 1600,
        2000, 2500, 3150, 4000, 5000, 6300, 8000, 10000, 20000, 48000,
    ]
)
# Ear Reference Point to Drum Reference Point (ERP-DRP) transfer function,
# Table 14A/P.58, sect 6.2. NB negative of table since defined other way round.

ITU_ERP_DRP = np.array(
    [
        0.0, 0.0, 0.0, 0.0, 0.0, 0.3, 0.2, 0.5, 0.6, 0.7, 1.1, 1.7, 2.6, 4.2,
        6.5, 9.4, 10.3, 6.6, 3.2, 3.3, 16, 14.4, 14.4, 14.4
    ]
)
# fmt: on

# Ideal pre-emphasis (starts off as SII 1997), then rescaled for Moore et al. 2008
# NB last two are -15dB/oct before rescaling below
# (Moore et al 2008 E&H paper suggests that shape would be better as -7.5
# dB/oct, at least up to 8, and -13 dB/oct above there.)

GEN_NOISE_HZ = np.array(
    [0, 100, 200, 450, 550, 707, 1000, 1414, 2000, 2828, 4000, 5656, 8000, 16e3, 32e3]
)
EMPHASIS = np.array(
    [0, 0.0, 0.0, 0, -0.5, -4.5, -9.0, -13.5, -18, -22.5, -27, -31.5, -36.0, -51, -66]
) * (7.5 / 9)


def read_gtf_file(gtf_file: str) -> Dict:
    """Read a gammatone filterbank file.

    List data is converted into numpy arrays.

    """

    # Fix filename if necessary
    dirname = os.path.dirname(os.path.abspath(__file__))
    gtf_file = os.path.join(dirname, gtf_file)
    with open(gtf_file, "r", encoding="utf-8") as fp:
        data = json.load(fp)
    for key in data:
        if isinstance(data[key], list):
            data[key] = np.array(data[key])
    return data


def firwin2(
    n_taps: int,
    frequencies: np.ndarray,
    filter_gains: np.ndarray,
    window: Union[None, str, tuple] = None,
    antisymmetric: Optional[bool] = None,
):  # pylint: disable=W0613
    """FIR filter design using the window method.

    Partial implementation of scipy firwin2 but using our own MATLAB-derived fir2.

    Args:
        n_taps (int): The number of taps in the FIR filter.
        frequencies (ndarray): The frequency sampling points. 0.0 to 1.0 with 1.0 being Nyquist.
        filter_gains (ndarray): The filter gains at the frequency sampling points.
        window (string or (string, float), optional): See scipy.firwin2 (default: (None))
        antisymmetric (bool, optional): Unused but present to main compatability
            with scipy firwin2.

    Returns:
        ndarray:  The filter coefficients of the FIR filter, as a 1-D array of length n.

    """
    window_shape = None
    window_type = None
    if isinstance(window, tuple):
        window_type, window_param = window if window is not None else (None, 0)
    else:
        window_type = window

    order = n_taps - 1

    if window_type == "kaiser":
        window_shape = scipy.signal.kaiser(n_taps, window_param)

    if window_shape is None:
        filter_coef, _ = fir2(order, frequencies, filter_gains)
    else:
        filter_coef, _ = fir2(order, frequencies, filter_gains, window_shape)

    return filter_coef


def fir2(
    filter_length: int,
    frequencies: np.ndarray,
    filter_gains: np.ndarray,
    n_interpolate: Optional[int] = None,
) -> Tuple[np.ndarray, int]:
    """FIR arbitrary shape filter design using the frequency sampling method.

    Translation of MATLAB fir2.

    Args:
        filter_length (int): Order
        frequencies (ndarray): The frequency sampling points (0 < frequencies < 1) where 1 is Nyquist rate.
                        First and last elements must be 0 and 1 respectively.
        filter_gains (ndarray): The filter gains at the frequency sampling points.
        n_interpolate (int, optional): Number of points for freq response interpolation
            (default: max(smallest power of 2 greater than nn, 512))

    Returns:
        np.ndarray: nn + 1 filter coefficients, 1

    """
    # Work with filter length instead of filter order
    filter_length += 1

    if n_interpolate is None:
        wind = scipy.signal.hamming(filter_length)
    else:
        wind = n_interpolate
    n_interpolate = (
        2 ** np.ceil(math.log(filter_length) / math.log(2))
        if filter_length >= 1024
        else 512
    )

    lap = np.fix(n_interpolate / 25).astype(int)

    nbrk = max(len(frequencies), len(filter_gains))

    frequencies[0] = 0
    frequencies[nbrk - 1] = 1

    H = np.zeros(n_interpolate + 1)
    nint = nbrk - 1
    df = np.diff(frequencies, n=1)

    n_interpolate += 1
    nb = 0
    H[0] = filter_gains[0]

    for i in np.arange(nint):
        if df[i] == 0:
            nb = int(np.ceil(nb - lap / 2))
            ne: int = nb + lap - 1
        else:
            ne = int(np.fix(frequencies[i + 1] * n_interpolate)) - 1

        j = np.arange(nb, ne + 1)
        inc: Union[float, np.ndarray] = 0.0 if nb == ne else (j - nb) / (ne - nb)
        H[nb : (ne + 1)] = inc * filter_gains[i + 1] + (1 - inc) * filter_gains[i]
        nb = ne + 1

    dt = 0.5 * (filter_length - 1)
    rad = -dt * 1j * math.pi * np.arange(0, n_interpolate) / (n_interpolate - 1)
    H = H * np.exp(rad)

    H = np.concatenate((H, H[n_interpolate - 2 : 0 : -1].conj()))
    ht = np.real(np.fft.ifft(H))

    b = ht[0:filter_length] * wind

    return b, 1


def gen_tone(
    freq: float, duration: float, sample_frequency: float = 44100.0, level: float = 0.0
) -> np.ndarray:
    """Generate a pure tone.

    Args:
        freq (float): Frequency of tone in Hz.
        duration (float): Duration of tone in seconds.
        sample_frequency (float, optional): Sample rate of generated tone in Hz. (default: 44100)
        level (float, optional): Level of tone in dB SPL. (default: 0)

    Returns:
        np.ndarray
    """
    return (
        1.4142
        * np.power(10, (0.05 * level))
        * np.sin(
            2
            * np.pi
            * freq
            * np.arange(1, duration * sample_frequency + 1)
            / sample_frequency
        )
    )


def gen_eh2008_speech_noise(
    duration: float,
    sample_frequency: float = 44100.0,
    level: Optional[float] = None,
    supplied_b: Optional[np.ndarray] = None,
) -> np.ndarray:
    """Generate speech shaped noise.

    Start with white noise and re-shape to ideal SII, ie flat to 500 Hz, and sloping -9db/oct beyond that.

    Slightly different shape from SII stylised same as EarHEar 2008 paper, Moore et al.

    Args:
        duration (float): Duration of signal in seconds
        sample_frequency (float): Sampling rate
        level (float, optional): Normalise to level dB if present
        supplied_b (ndarray, optional): High-pass filter (default: uses built-in pre-emphasis filter)

    Returns:
        ndarray: Noise signal

    """
    sample_frequency = int(sample_frequency)
    n_samples = int(duration * sample_frequency)

    # this rescales so that we get -7.5 dB/oct up to 8kHz, and -13 dB/oct above that
    norm_freq = GEN_NOISE_HZ / (sample_frequency / 2)
    last_f_idx = np.max(np.where(norm_freq < 1))
    norm_freq = np.append(norm_freq[0 : last_f_idx + 1], 1)

    # -9 dB/oct constant slope
    emph_nyq = EMPHASIS[last_f_idx] + 9 * np.log10(norm_freq[last_f_idx]) / np.log10(2)
    norm_emph = np.append(EMPHASIS[0 : last_f_idx + 1], emph_nyq)
    m = np.exp(np.log(10) * norm_emph / 20)

    # Create type II filter with 10 msec window and even number of taps
    n_taps = int(2 * np.ceil(10 * (sample_frequency / 2000))) + 1
    b = (
        supplied_b
        if supplied_b is not None
        else firwin2(n_taps, norm_freq, m, window="hamming", antisymmetric=False)
    )

    # white noise, 0 DC
    n_burst = np.random.random((1, n_samples + len(b))) - 0.5

    # remove low-freq noise that may bias RMS estimate, -33dB at 50 Hz
    eh2008_nse = scipy.signal.lfilter(b, 1, n_burst)

    # high-pass filter to remove low freqs (will be 2-pass with filtfilt)
    high_pass_filter = scipy.signal.ellip(
        3, 0.1, 50, 100 / (sample_frequency / 2), "high"
    )
    padlen = 3 * (max(len(high_pass_filter[1]), len(high_pass_filter[0])) - 1)
    eh2008_nse = scipy.signal.filtfilt(
        *high_pass_filter, eh2008_nse, padlen=padlen
    ).flatten()

    # this introduces a delay so remove it, ie time-ADVANCE audio
    # compensating shift to time-align all filter outputs
    delay_shift = int(np.floor(len(b) / 2))
    valid_len = int(np.size(eh2008_nse) - delay_shift)  # _advance_ filter outputs
    # time advance
    eh2008_nse[0:valid_len] = eh2008_nse[delay_shift:]
    eh2008_nse = eh2008_nse[0:n_samples]

    if level is not None:
        eh2008_nse = (
            eh2008_nse
            * np.power(10, 0.05 * level)
            / np.sqrt(np.sum(np.power(eh2008_nse, 2)) / len(eh2008_nse))
        )

    return eh2008_nse


def generate_key_percent(
    signal: np.ndarray,
    threshold_db: float,
    window_length: int,
    percent_to_track: Optional[float] = None,
) -> Tuple[np.ndarray, float]:
    """Generate key percent.
    Locates frames above some energy threshold or tracks a certain percentage
    of frames. To track a certain percentage of frames in order to get measure
    of rms, adaptively sets threshold after looking at histogram of whole recording

    Args:
        signal (ndarray): The signal to analyse.
        threshold_db (float): fixed energy threshold (dB).
        window_length (int): length of window in samples.
        percent_to_track (float, optional): Track a percentage of frames (default: {None}).

    Raises:
        ValueError: percent_to_track is set too high.

    Returns:
        (tuple): containig
        - key (ndarray): The key array of indices of samples used in rms calculation.
        - used_threshold_db (float): Root Mean Squared threshold.

            and the threshold used to get a more accurate rms calculation
    """
    window_length = int(window_length)
    signal = signal.flatten()
    if window_length != math.floor(
        window_length
    ):  # whoops on fractional indexing: 7-March 2002
        window_length = math.floor(window_length)
        logging.warning(f"Window length must be integer: now {window_length}")

    signal_length = len(signal)

    expected = threshold_db
    # new Dec 2003. Possibly track percentage of frames rather than fixed threshold
    if percent_to_track is not None:
        logging.info("tracking %s percentage of frames", percent_to_track)
    else:
        logging.info("tracking fixed threshold")

    # put floor into histogram distribution
    non_zero = np.power(10, (expected - 30) / 10)

    n_frames = -1
    total_frames = math.floor(signal_length / window_length)
    every_db = np.zeros(total_frames)

    for ix in np.arange(
        0, window_length * total_frames - 1, window_length
    ):  # pylint: disable=invalid-name
        n_frames += 1
        this_sum = np.sum(
            np.power(signal[ix : (ix + window_length)].astype("float"), 2)
        )
        every_db[n_frames] = 10 * np.log10(non_zero + this_sum / window_length)
    n_frames += 1

    # from now on save only those analysed
    every_db = every_db[:n_frames]

    # Bec 2003, was 100 to give about a 0.5 dB quantising of levels
    n_bins, levels = np.histogram(every_db, 140)
    if percent_to_track is not None:
        # min number of bins to use
        inactive_bins = (100 - percent_to_track) * n_frames / 100
        n_levels = len(levels)
        inactive_ix = 0
        ix_count = 0
        for ix in np.arange(0, n_levels, 1):  # pylint: disable=invalid-name
            inactive_ix = inactive_ix + n_bins[ix]
            if inactive_ix > inactive_bins:
                break
            ix_count += 1
        if ix == 1:
            logging.warning("Counted every bin.........")
        elif ix == n_levels:
            raise ValueError("Generate_key_percent: no levels to count")
        expected = levels[max(1, ix_count)]

    # set new threshold conservatively to include more bins than desired
    used_threshold_db = expected

    # histogram should produce a two-peaked curve: thresh should be set in valley
    # between the two peaks, and set threshold a bit above that,
    # as it heads for main peak
    # FixMe : Could Otsu's method (from image processing) be used here? https://en.wikipedia.org/wiki/Otsu's_method
    frame_index = np.nonzero(every_db >= expected)[0]
    valid_frames = len(frame_index)
    key = np.zeros((1, valid_frames * window_length))[0]

    # convert frame numbers into indices for sig
    for ix in np.arange(valid_frames):  # pylint: disable=invalid-name
        meas_span = np.arange(
            (frame_index[ix] * window_length), (frame_index[ix] + 1) * window_length
        )
        key_span = np.arange(((ix) * window_length), (ix + 1) * window_length, 1)
        key[key_span] = meas_span
        key = key.flatten()

    return key, used_threshold_db


def measure_rms(
    signal: np.ndarray,
    sample_frequency: float,
    db_rel_rms: float,
    percent_to_track: Optional[float] = None,
) -> tuple:
    """Measure Root Mean Square.

    A sophisticated method of measuring RMS in a file. It splits the signal up into
    short windows, performs  a histogram of levels, calculates an approximate RMS,
    and then uses that RMS to calculate a threshold level in the histogram and then
    re-measures the RMS only using those durations whose individual RMS exceed that
    threshold.

    Args:
        signal (ndarray): the signal of which to measure the Root Mean Square.
        sample_frequency (float): sampling frequency.
        db_rel_rms (float): threshold for frames to track.
        percent_to_track (float, optional): track percentage of frames,
            rather than threshold (default: {None})
    Returns:
        (tuple): tuple containing
        - rms (float): overall calculated rms (linear)
        - key (ndarray): "key" array of indices of samples used in rms calculation
        - rel_db_thresh (float): fixed threshold value of -12 dB
        - active (float): proportion of values used in rms calculation
    """
    sample_frequency = int(sample_frequency)
    # first RMS is of all signal.
    first_stage_rms = np.sqrt(np.sum(np.power(signal, 2) / len(signal)))
    # use this RMS to generate key threshold to get more accurate RMS
    key_thr_db = max(20 * np.log10(first_stage_rms) + db_rel_rms, -80)

    # move key_thr_db to account for noise less peakier than signal
    key, used_thr_db = generate_key_percent(
        signal,
        key_thr_db,
        round(WIN_SECS * sample_frequency),
        percent_to_track=percent_to_track,
    )

    idx = key.astype(int)  # move into generate_key_percent
    # statistic to be reported later, BUT save for later
    # (for independent==1 loop where it sets a target for rms measure)
    active = 100 * len(key) / len(signal)
    rms = np.sqrt(np.sum(np.power(signal[idx], 2)) / len(key))
    rel_db_thresh = used_thr_db - 20 * np.log10(rms)

    return rms, idx, rel_db_thresh, active


def pad(signal, length):
    """Zero pad signal to required length.

    Assumes required length is not less than input length.
    """
    assert length >= signal.shape[0]
    return np.pad(
        signal, [(0, length - signal.shape[0])] + [(0, 0)] * (len(signal.shape) - 1)
    )


def read_signal(
    filename: str,
    offset: int = 0,
    nsamples: int = -1,
    nchannels: int = 0,
    offset_is_samples: bool = False,
) -> np.ndarray:
    """Read a wavefile and return as numpy array of floats.

    Args:
        filename (string): Name of file to read
        offset (int, optional): Offset in samples or seconds (from start). Defaults to 0.
        nsamples (int): Number of samples.
        nchannels (int): expected number of channel (default: 0 = any number OK)
        offset_is_samples (bool): measurement units for offset (default: False)

    Returns:
        np.ndarray: audio signal
    """
    try:
        wave_file = SoundFile(filename)
    except Exception as e:
        # Ensure incorrect error (24 bit) is not generated
        raise Exception(f"Unable to read {filename}.") from e

    if nchannels not in (0, wave_file.channels):
        raise Exception(
            f"Wav file ({filename}) was expected to have {nchannels} channels."
        )

    if not offset_is_samples:  # Default behaviour
        offset = int(offset * wave_file.samplerate)

    if offset != 0:
        wave_file.seek(offset)

    signal = wave_file.read(frames=nsamples)

    if wave_file.samplerate != MSBG_FS:
        signal = scipy.signal.resample(
            signal, int(MSBG_FS * signal.shape[0] / wave_file.samplerate)
        )

    return signal


def write_signal(
    filename: str,
    signal: np.ndarray,
    sample_frequency: float,
    floating_point: bool = True,
) -> None:
    """Write a signal as fixed or floating point wav file.

    Args:
        filename (str): name of file in to write to.
        signal (ndarray): signal to write.
        sample_frequency (float): sampling frequency.
        floating_point (bool): write as floating point else an ints (default: True).
    """

    if sample_frequency != MSBG_FS:
        logging.warning(
            f"Sampling rate mismatch: {filename} with sample frequency = {sample_frequency}."
        )
        # raise ValueError("Sampling rate mismatch")

    if floating_point is False:
        if TEST_NBITS == 16:
            subtype = "PCM_16"
            # If signal is float and we want int16
            signal *= 32768
            signal = signal.astype(np.dtype("int16"))
            assert np.max(signal) <= 32767 and np.min(signal) >= -32768
        elif TEST_NBITS == 24:
            subtype = "PCM_24"
    else:
        subtype = "FLOAT"

    soundfile.write(filename, signal, sample_frequency, subtype=subtype)



""" Run the dummy enhancement. """
import json
import logging
import pathlib

import hydra
import numpy as np
from evaluate import make_scene_listener_list
from omegaconf import DictConfig
from scipy.io import wavfile
from tqdm import tqdm

logger = logging.getLogger(__name__)


@hydra.main(config_path=".", config_name="config")
def enhance(cfg: DictConfig) -> None:
    """Run the dummy enhancement."""

    enhanced_folder = pathlib.Path("enhanced_signals")
    enhanced_folder.mkdir(parents=True, exist_ok=True)

    with open(cfg.path.scenes_listeners_file, "r", encoding="utf-8") as fp:
        scenes_listeners = json.load(fp)

    with open(cfg.path.listeners_file, "r", encoding="utf-8") as fp:
        listener_audiograms = json.load(fp)  # noqa: F841

    # Make list of all scene listener pairs that will be run
    scene_listener_pairs = make_scene_listener_list(
        scenes_listeners, cfg.evaluate.small_test
    )

    for scene, listener in tqdm(scene_listener_pairs):
        sample_freq, signal = wavfile.read(
            pathlib.Path(cfg.path.scenes_folder) / f"{scene}_mix_CH1.wav"
        )

        # Convert to 32-bit floating point scaled between -1 and 1
        signal = (signal / 32768.0).astype(np.float32)

        # # Audiograms can read like this, but they are not needed for the baseline
        #
        # cfs = np.array(listener_audiograms[listener]["audiogram_cfs"])
        #
        # audiogram_left = np.array(
        #    listener_audiograms[listener]["audiogram_levels_l"]
        # )
        # audiogram_right = np.array(
        #    listener_audiograms[listener]["audiogram_levels_r"]
        # )

        # Baseline just reads the signal from the front microphone pair
        # and write it out as the enhanced signal
        #

        wavfile.write(
            enhanced_folder / f"{scene}_{listener}_enhanced.wav", sample_freq, signal
        )


if __name__ == "__main__":
    enhance()



import csv
import json
import logging
import os
import random

import hydra
import numpy as np
import soundfile as sf
from librosa import resample
from omegaconf import DictConfig
from tqdm import tqdm

from clarity.evaluator.msbg.audiogram import Audiogram
from clarity.evaluator.msbg.msbg import Ear
from clarity.evaluator.msbg.msbg_utils import MSBG_FS, pad, read_signal, write_signal

logger = logging.getLogger(__name__)

targ_fs = 16000


def run_data_split(cfg, track):
    os.makedirs(os.path.join(cfg.path.exp_folder, "data_split"), exist_ok=True)
    scene_train_json = os.path.join(
        cfg.path.exp_folder, "data_split", f"scene_train_list{track}.json"
    )
    scene_dev_json = os.path.join(
        cfg.path.exp_folder, "data_split", f"scene_dev_list{track}.json"
    )
    if os.path.isfile(scene_train_json) and os.path.isfile(scene_dev_json):
        logger.info("Train set and dev set lists exist...")
        return

    scenes_dict = json.load(
        open(
            os.path.join(
                cfg.path.root,
                "clarity_CPC1_data_train/metadata",
                f"CPC1.{'train'+track}.json",
            )
        )
    )
    scene_list = []
    for item in scenes_dict:
        scene_list.append(item["scene"])
    scene_list = list(set(scene_list))
    scene_dev_list = random.sample(
        scene_list, int(np.floor(len(scene_list) * cfg.dev_percent))
    )
    scene_train_list = list(set(scene_list) - set(scene_dev_list))

    with open(scene_train_json, "w") as f:
        json.dump(scene_train_list, f)
    with open(scene_dev_json, "w") as f:
        json.dump(scene_dev_list, f)


def listen(ear, signal, audiogram_l, audiogram_r):
    """
    Generate MSBG processed signal
    :param ear: MSBG ear
    :param wav: binaural signal
    :return: binaural signal
    """
    ear.set_audiogram(audiogram_l)
    out_l = ear.process(signal[:, 0])

    ear.set_audiogram(audiogram_r)
    out_r = ear.process(signal[:, 1])
    if len(out_l[0]) != len(out_r[0]):
        diff = len(out_l[0]) - len(out_r[0])
        if diff > 0:
            out_r[0] = np.flipud(pad(np.flipud(out_r[0]), len(out_l[0])))
        else:
            out_l[0] = np.flipud(pad(np.flipud(out_l[0]), len(out_r[0])))
    return np.concatenate([out_l, out_r]).T


def run_msbg_simulation(cfg, track):
    for split in ["train", "test"]:
        dataset = split + track
        dataset_folder = os.path.join(cfg.path.root, "clarity_CPC1_data_" + split)
        output_path = os.path.join(dataset_folder, "clarity_data/HA_outputs", dataset)
        scenes = json.load(
            open(os.path.join(dataset_folder, "metadata", f"CPC1.{dataset}.json"))
        )
        if split == "train":
            listener_audiograms = json.load(
                open(
                    os.path.join(
                        dataset_folder, "metadata", "listeners.CPC1_train.json"
                    )
                )
            )
        else:
            listener_audiograms = json.load(
                open(
                    os.path.join(dataset_folder, "metadata", "listeners.CPC1_all.json")
                )
            )

        # initialize ear
        ear = Ear(**cfg["MSBGEar"])

        for scene_dict in tqdm(scenes):
            scene = scene_dict["scene"]
            listener = scene_dict["listener"]
            system = scene_dict["system"]
            signal_file = os.path.join(output_path, f"{scene}_{listener}_{system}.wav")

            # signals to write
            outfile_stem = f"{output_path}/{scene}_{listener}_{system}"
            signal_files_to_write = [
                f"{outfile_stem}_HL-output.wav",
            ]
            # if all signals to write exist, pass
            if all([os.path.isfile(f) for f in signal_files_to_write]):
                continue
            signal = read_signal(signal_file)

            # retrieve audiograms
            cfs = np.array(listener_audiograms[listener]["audiogram_cfs"])
            audiogram_left = np.array(
                listener_audiograms[listener]["audiogram_levels_l"]
            )
            left_audiogram = Audiogram(cfs=cfs, levels=audiogram_left)
            audiogram_right = np.array(
                listener_audiograms[listener]["audiogram_levels_r"]
            )
            right_audiogram = Audiogram(cfs=cfs, levels=audiogram_right)

            signals_to_write = [
                listen(ear, signal, left_audiogram, right_audiogram),
            ]
            for i in range(len(signals_to_write)):
                write_signal(
                    signal_files_to_write[i],
                    signals_to_write[i],
                    MSBG_FS,
                    floating_point=True,
                )


def generate_data_split(
    orig_data_json,
    orig_signal_folder,
    target_data_folder,
    data_split,
    split_data_list,
    if_msbg=False,
    if_ref=False,
):

    with open(orig_data_json, "r") as f:
        all_data_list = json.load(f)

    left_tgt_signal_folder = os.path.join(
        target_data_folder, orig_signal_folder.split("/")[-2] + "_left"
    )
    right_tgt_signal_folder = os.path.join(
        target_data_folder, orig_signal_folder.split("/")[-2] + "_right"
    )
    if if_msbg:
        left_tgt_signal_folder = left_tgt_signal_folder + "_msbg"
        right_tgt_signal_folder = right_tgt_signal_folder + "_msbg"
    if if_ref:
        left_tgt_signal_folder = left_tgt_signal_folder + "_ref"
        right_tgt_signal_folder = right_tgt_signal_folder + "_ref"
    os.makedirs(left_tgt_signal_folder, exist_ok=True)
    os.makedirs(right_tgt_signal_folder, exist_ok=True)

    csv_lines_left = [["ID", "duration", "wav", "spk_id", "wrd"]]
    csv_lines_right = [["ID", "duration", "wav", "spk_id", "wrd"]]
    csv_lines_binaural = [["ID", "duration", "wav", "spk_id", "wrd"]]

    left_csvfile = "left_" + data_split
    right_csvfile = "right_" + data_split
    binaural_csvfile = "binaural_" + data_split
    if if_msbg:
        left_csvfile += "_msbg"
        right_csvfile += "_msbg"
        binaural_csvfile += "_msbg"
    if if_ref:
        left_csvfile += "_ref"
        right_csvfile += "_ref"
        binaural_csvfile += "_ref"
    left_csvfile = os.path.join(target_data_folder, left_csvfile + ".csv")
    right_csvfile = os.path.join(target_data_folder, right_csvfile + ".csv")
    binaural_csvfile = os.path.join(target_data_folder, binaural_csvfile + ".csv")

    for item in tqdm(all_data_list):
        snt_id = item["signal"]
        if snt_id.split("_")[0] in split_data_list or data_split == "test":
            if if_msbg:
                wav_file = os.path.join(orig_signal_folder, snt_id + "_HL-output.wav")
            elif if_ref:
                wav_file = os.path.join(
                    orig_signal_folder,
                    "../..",
                    "scenes",
                    snt_id.split("_")[0] + "_target_anechoic.wav",
                )
            else:
                wav_file = os.path.join(orig_signal_folder, snt_id + ".wav")
            wav_file_left = os.path.join(left_tgt_signal_folder, snt_id + ".wav")
            wav_file_right = os.path.join(right_tgt_signal_folder, snt_id + ".wav")
            if os.path.isfile(wav_file_left) and os.path.isfile(wav_file_right):
                continue

            spk_id = item["listener"]  # should be listener_id
            wrds = item["prompt"].upper()

            utt, orig_fs = sf.read(wav_file)
            utt_16k = resample(np.array(utt).transpose(), orig_fs, targ_fs).transpose()
            duration = (
                len(utt_16k[2 * targ_fs :, 0]) / targ_fs
            )  # Get rid of the first two seconds, as there is no speech
            sf.write(wav_file_left, utt_16k[2 * targ_fs :, 0], targ_fs)
            sf.write(wav_file_right, utt_16k[2 * targ_fs :, 1], targ_fs)

            csv_lines_left.append([snt_id, str(duration), wav_file_left, spk_id, wrds])
            csv_lines_right.append(
                [snt_id, str(duration), wav_file_right, spk_id, wrds]
            )
            csv_lines_binaural.append(
                ["left_" + snt_id, str(duration), wav_file_left, spk_id, wrds]
            )
            csv_lines_binaural.append(
                ["right_" + snt_id, str(duration), wav_file_right, spk_id, wrds]
            )

    with open(left_csvfile, mode="w") as csv_f:
        csv_writer = csv.writer(
            csv_f, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL
        )
        for line in csv_lines_left:
            csv_writer.writerow(line)

    with open(right_csvfile, mode="w") as csv_f:
        csv_writer = csv.writer(
            csv_f, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL
        )
        for line in csv_lines_right:
            csv_writer.writerow(line)

    with open(binaural_csvfile, mode="w") as csv_f:
        csv_writer = csv.writer(
            csv_f, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL
        )
        for line in csv_lines_binaural:
            csv_writer.writerow(line)


def run_signal_generation_train(cfg, track):
    train_json_path = os.path.join(
        cfg.path.cpc1_train_data, f"metadata/CPC1.{'train'+track}.json"
    )
    train_signal_folder = os.path.join(
        cfg.path.cpc1_train_data, f"clarity_data/HA_outputs/{'train'+track}/"
    )
    target_folder = os.path.join(cfg.path.exp_folder, "cpc1_asr_data" + track)
    lists_to_generate = [
        os.path.join(
            cfg.path.exp_folder, "data_split", f"scene_train_list{track}.json"
        ),
        os.path.join(cfg.path.exp_folder, "data_split", f"scene_dev_list{track}.json"),
    ]

    datasets_to_generate = []
    for i in range(len(lists_to_generate)):
        with open(lists_to_generate[i], "r") as f:
            datasets_to_generate.append(json.load(f))
            f.close()

    for i in range(len(lists_to_generate)):
        # generate_data_split(
        #     train_json_path,
        #     train_signal_folder,
        #     target_folder,
        #     os.path.basename(lists_to_generate[i]).split("_")[1],
        #     datasets_to_generate[i],
        # )

        generate_data_split(
            train_json_path,
            train_signal_folder,
            target_folder,
            os.path.basename(lists_to_generate[i]).split("_")[1],
            datasets_to_generate[i],
            if_msbg=True,
        )

        generate_data_split(
            train_json_path,
            train_signal_folder,
            target_folder,
            os.path.basename(lists_to_generate[i]).split("_")[1],
            datasets_to_generate[i],
            if_ref=True,
        )


def run_signal_generation_test(cfg, track):
    test_json_path = os.path.join(
        cfg.path.cpc1_test_data, f"metadata/CPC1.{'test'+track}.json"
    )
    test_signal_folder = os.path.join(
        cfg.path.cpc1_test_data, f"clarity_data/HA_outputs/{'test'+track}/"
    )
    target_folder = os.path.join(cfg.path.exp_folder, "cpc1_asr_data" + track)

    # generate_data_split(
    #     test_json_path, test_signal_folder, target_folder, "test", [],
    # )

    generate_data_split(
        test_json_path, test_signal_folder, target_folder, "test", [], if_msbg=True
    )

    generate_data_split(
        test_json_path, test_signal_folder, target_folder, "test", [], if_ref=True
    )


@hydra.main(config_path=".", config_name="config")
def run(cfg: DictConfig) -> None:
    if cfg.cpc1_track == "open":
        track = "_indep"
    elif cfg.cpc1_track == "closed":
        track = ""
    else:
        logger.error("cpc1_track has to be closed or open")

    logger.info("Split all training data into train set and dev set.")
    run_data_split(cfg, track)
    logger.info("Running MSGB simulation.")
    run_msbg_simulation(cfg, track)
    logger.info("Begin generating signals for ASR for cpc1_train_data.")
    run_signal_generation_train(cfg, track)
    logger.info("Begin generating signals for ASR for cpc1_test_data.")
    run_signal_generation_test(cfg, track)


if __name__ == "__main__":
    run()



import json
import logging
import os

import hydra
import numpy as np
from omegaconf import DictConfig
from scipy.io import wavfile
from tqdm import tqdm

from clarity.enhancer.compressor import Compressor
from clarity.enhancer.nalr import NALR

logger = logging.getLogger(__name__)


@hydra.main(config_path=".", config_name="config")
def enhance(cfg: DictConfig) -> None:
    enhanced_folder = os.path.join(cfg.path.exp_folder, "enhanced_signals")
    os.makedirs(enhanced_folder, exist_ok=True)
    scenes_listeners = json.load(open(cfg.path.scenes_listeners_file))
    listener_audiograms = json.load(open(cfg.path.listeners_file))

    enhancer = NALR(**cfg.nalr)
    compressor = Compressor(**cfg.compressor)

    for scene in tqdm(scenes_listeners):
        for listener in scenes_listeners[scene]:
            # retrieve audiograms
            cfs = np.array(listener_audiograms[listener]["audiogram_cfs"])
            audiogram_left = np.array(
                listener_audiograms[listener]["audiogram_levels_l"]
            )
            audiogram_right = np.array(
                listener_audiograms[listener]["audiogram_levels_r"]
            )

            fs, signal = wavfile.read(
                os.path.join(cfg.path.scenes_folder, f"{scene}_mix_CH1.wav")
            )
            signal = signal / 32768.0
            assert fs == cfg.nalr.fs
            nalr_fir, _ = enhancer.build(audiogram_left, cfs)
            out_l = enhancer.apply(nalr_fir, signal[:, 0])

            nalr_fir, _ = enhancer.build(audiogram_right, cfs)
            out_r = enhancer.apply(nalr_fir, signal[:, 1])

            out_l, _, _ = compressor.process(out_l)
            out_r, _, _ = compressor.process(out_r)
            enhanced = np.stack([out_l, out_r], axis=1)
            filename = f"{scene}_{listener}_HA-output.wav"

            if cfg.soft_clip:
                enhanced = np.tanh(enhanced)
            n_clipped = np.sum(np.abs(enhanced) > 1.0)
            if n_clipped > 0:
                logger.warning(f"Writing {filename}: {n_clipped} samples clipped")
            np.clip(enhanced, -1.0, 1.0, out=enhanced)
            signal_16 = (32768.0 * enhanced).astype(np.int16)
            wavfile.write(os.path.join(enhanced_folder, filename), fs, signal_16)


if __name__ == "__main__":
    enhance()



import csv
import hashlib
import json
import logging
import os

import hydra
import numpy as np
from omegaconf import DictConfig
from scipy.io import wavfile
from tqdm import tqdm

from clarity.evaluator.haspi import haspi_v2_be

logger = logging.getLogger(__name__)


def read_csv_scores(file):
    score_dict = {}
    with open(file, "r") as f:
        reader = csv.reader(f)
        _ = next(reader)
        for row in reader:
            score_dict[row[0] + "_" + row[1]] = float(row[2])
    return score_dict


@hydra.main(config_path=".", config_name="config")
def run_calculate_SI(cfg: DictConfig) -> None:
    scenes_listeners = json.load(open(cfg.path.scenes_listeners_file))
    listener_audiograms = json.load(open(cfg.path.listeners_file))
    os.makedirs(cfg.path.exp_folder, exist_ok=True)

    enhanced_folder = os.path.join(cfg.path.exp_folder, "enhanced_signals")

    if cfg.evaluate.cal_unprocessed_si:
        unproc_si_file = os.path.join(cfg.path.exp_folder, "si_unproc.csv")
        unproc_csv_lines = [["scene", "listener", "haspi"]]
        if os.path.exists(unproc_si_file):
            score_dict = read_csv_scores(unproc_si_file)
            ave_score = np.mean(list(score_dict.values()))
            logger.info(
                "si_unproc.csv exists, and the average HASPI scores for unprocessed scenes is {:.4f}".format(
                    ave_score
                )
            )

    si_file = os.path.join(cfg.path.exp_folder, "si.csv")
    csv_lines = [["scene", "listener", "haspi"]]
    if os.path.exists(si_file):
        score_dict = read_csv_scores(si_file)
        ave_score = np.mean(list(score_dict.values()))
        logger.info(
            "si.csv exists, and the average HASPI scores is {:.4f}".format(ave_score)
        )
        return

    for scene in tqdm(scenes_listeners):
        for listener in scenes_listeners[scene]:
            logger.info(f"Running SI calculation: scene {scene}, listener {listener}")
            if cfg.evaluate.set_random_seed:
                scene_md5 = int(hashlib.md5(scene.encode("utf-8")).hexdigest(), 16) % (
                    10**8
                )
                np.random.seed(scene_md5)

            # retrieve audiograms
            cfs = np.array(listener_audiograms[listener]["audiogram_cfs"])
            audiogram_left = np.array(
                listener_audiograms[listener]["audiogram_levels_l"]
            )
            audiogram_right = np.array(
                listener_audiograms[listener]["audiogram_levels_r"]
            )

            # retrieve signals
            fs_proc, proc = wavfile.read(
                os.path.join(enhanced_folder, f"{scene}_{listener}_HA-output.wav")
            )

            fs_ref_anechoic, ref_anechoic = wavfile.read(
                os.path.join(cfg.path.scenes_folder, f"{scene}_target_anechoic_CH1.wav")
            )

            fs_ref_target, ref_target = wavfile.read(
                os.path.join(cfg.path.scenes_folder, f"{scene}_target_CH1.wav")
            )

            assert fs_ref_anechoic == fs_ref_target == fs_proc

            proc = proc / 32768.0
            ref_anechoic = ref_anechoic / 32768.0
            ref_target = ref_target / 32768.0

            rms_target = np.mean(ref_target**2, axis=0) ** 0.5
            rms_anechoic = np.mean(ref_anechoic**2, axis=0) ** 0.5
            ref = ref_anechoic * rms_target / rms_anechoic

            si = haspi_v2_be(
                xl=ref[:, 0],
                xr=ref[:, 1],
                yl=proc[:, 0],
                yr=proc[:, 1],
                fs_signal=fs_ref_anechoic,
                audiogram_l=audiogram_left,
                audiogram_r=audiogram_right,
                audiogram_cfs=cfs,
            )
            logger.info(f"The HASPI score is {si}")
            csv_lines.append([scene, listener, str(si)])

            if cfg.evaluate.cal_unprocessed_si:
                if cfg.evaluate.set_random_seed:
                    scene_md5 = int(
                        hashlib.md5(scene.encode("utf-8")).hexdigest(), 16
                    ) % (10**8)
                    np.random.seed(scene_md5)

                fs_unproc, unproc = wavfile.read(
                    os.path.join(cfg.path.scenes_folder, f"{scene}_mix_CH1.wav")
                )
                unproc = unproc / 32768.0
                si_unproc = haspi_v2_be(
                    xl=ref[:, 0],
                    xr=ref[:, 1],
                    yl=unproc[:, 0],
                    yr=unproc[:, 1],
                    fs_signal=fs_ref_anechoic,
                    audiogram_l=audiogram_left,
                    audiogram_r=audiogram_right,
                    audiogram_cfs=cfs,
                )
                logger.info(f"The unprocessed signal HASPI score is {si_unproc}")
                unproc_csv_lines.append([scene, listener, str(si_unproc)])

    with open(si_file, "w") as csv_f:
        csv_writer = csv.writer(
            csv_f, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL
        )
        for line in csv_lines:
            csv_writer.writerow(line)
    score_dict = read_csv_scores(si_file)
    ave_score = np.mean(list(score_dict.values()))
    logger.info(
        "si.csv exists, and the average HASPI scores is {:.4f}".format(ave_score)
    )

    if cfg.evaluate.cal_unprocessed_si:
        with open(unproc_si_file, "w") as csv_f:
            csv_writer = csv.writer(
                csv_f, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL
            )
            for line in unproc_csv_lines:
                csv_writer.writerow(line)
        score_dict = read_csv_scores(unproc_si_file)
        ave_score = np.mean(list(score_dict.values()))
        logger.info(
            "si_unproc.csv exists, and the average HASPI scores for unprocessed scenes is {:.4f}".format(
                ave_score
            )
        )


if __name__ == "__main__":
    run_calculate_SI()



import csv
import json
import logging
import os

import hydra
import numpy as np
from omegaconf import DictConfig
from scipy.signal import unit_impulse
from tqdm import tqdm

from clarity.evaluator.mbstoi.mbstoi import mbstoi
from clarity.evaluator.mbstoi.mbstoi_utils import find_delay_impulse
from clarity.evaluator.msbg.audiogram import Audiogram
from clarity.evaluator.msbg.msbg import Ear
from clarity.evaluator.msbg.msbg_utils import MSBG_FS, pad, read_signal, write_signal

logger = logging.getLogger(__name__)


def listen(ear, signal, audiogram_l, audiogram_r):
    """
    Generate MSBG processed signal
    :param ear: MSBG ear
    :param wav: binaural signal
    :return: binaural signal
    """
    ear.set_audiogram(audiogram_l)
    out_l = ear.process(signal[:, 0])

    ear.set_audiogram(audiogram_r)
    out_r = ear.process(signal[:, 1])
    if len(out_l[0]) != len(out_r[0]):
        diff = len(out_l[0]) - len(out_r[0])
        if diff > 0:
            out_r[0] = np.flipud(pad(np.flipud(out_r[0]), len(out_l[0])))
        else:
            out_l[0] = np.flipud(pad(np.flipud(out_l[0]), len(out_r[0])))
    return np.concatenate([out_l, out_r]).T


def run_HL_processing(cfg, path):
    output_path = os.path.join(path.exp_folder, "eval_signals")
    os.makedirs(output_path, exist_ok=True)
    scenes = json.load(open(path.scenes_file))
    listener_audiograms = json.load(open(path.listeners_file))
    enhanced_folder = path.scenes_folder

    # initialize ear
    ear = Ear(**cfg["MSBGEar"])

    for scene_dict in tqdm(scenes):
        scene = scene_dict["scene"]
        listener = scene_dict["listener"]
        system = scene_dict["system"]
        signal_file = os.path.join(enhanced_folder, f"{scene}_{listener}_{system}.wav")

        # signals to write
        outfile_stem = f"{output_path}/{scene}_{listener}_{system}"
        signal_files_to_write = [
            f"{output_path}/{scene}_flat0dB_HL-output.wav",
            f"{outfile_stem}_HL-output.wav",
            f"{outfile_stem}_HLddf-output.wav",
            f"{outfile_stem}_HL-mixoutput.wav",
        ]
        # if all signals to write exist, pass
        if all([os.path.isfile(f) for f in signal_files_to_write]):
            continue

        signal = read_signal(signal_file)
        mixture_signal = read_signal(
            os.path.join(path.ref_folder, f"{scene}_mixed_CH0.wav")
        )

        # retrieve audiograms
        cfs = np.array(listener_audiograms[listener]["audiogram_cfs"])
        audiogram_left = np.array(listener_audiograms[listener]["audiogram_levels_l"])
        left_audiogram = Audiogram(cfs=cfs, levels=audiogram_left)
        audiogram_right = np.array(listener_audiograms[listener]["audiogram_levels_r"])
        right_audiogram = Audiogram(cfs=cfs, levels=audiogram_right)

        # Create discrete delta function (DDF) signal for time alignment
        ddf_signal = np.zeros((np.shape(signal)))
        ddf_signal[:, 0] = unit_impulse(len(signal), int(MSBG_FS / 2))
        ddf_signal[:, 1] = unit_impulse(len(signal), int(MSBG_FS / 2))

        # Get flat-0dB ear audiograms
        flat0dB_audiogram = Audiogram(cfs=cfs, levels=np.zeros((np.shape(cfs))))

        signals_to_write = [
            listen(ear, ddf_signal, flat0dB_audiogram, flat0dB_audiogram),
            listen(ear, signal, left_audiogram, right_audiogram),
            listen(ear, ddf_signal, left_audiogram, right_audiogram),
            listen(ear, mixture_signal, left_audiogram, right_audiogram),
        ]
        for i in range(len(signals_to_write)):
            write_signal(
                signal_files_to_write[i],
                signals_to_write[i],
                MSBG_FS,
                floating_point=True,
            )


def run_calculate_SI(cfg, path) -> None:
    scenes = json.load(open(path.scenes_file))
    proc_folder = os.path.join(path.exp_folder, "eval_signals")
    sii_file = os.path.join(path.exp_folder, "sii.csv")
    csv_lines = [["signal_ID", "intelligibility_score"]]

    for scene_dict in tqdm(scenes):
        scene = scene_dict["scene"]
        listener = scene_dict["listener"]
        system = scene_dict["system"]
        proc = read_signal(
            os.path.join(proc_folder, f"{scene}_{listener}_{system}_HL-output.wav")
        )
        clean = read_signal(
            os.path.join(path.ref_folder, f"{scene}_target_anechoic.wav")
        )
        ddf = read_signal(
            os.path.join(proc_folder, f"{scene}_{listener}_{system}_HLddf-output.wav")
        )

        # Calculate channel-specific unit impulse delay due to HL model and audiograms
        delay = find_delay_impulse(ddf, initial_value=int(MSBG_FS / 2))
        if delay[0] != delay[1]:
            logging.info(f"Difference in delay of {delay[0] - delay[1]}.")
        maxdelay = int(np.max(delay))

        # Allow for value lower than 1000 samples in case of unimpaired hearing
        if maxdelay > 2000:
            logging.error("Error in delay calculation for signal time-alignment.")

        # Correct for delays by padding clean signals
        cleanpad = np.zeros((len(clean) + maxdelay, 2))
        procpad = np.zeros((len(clean) + maxdelay, 2))

        if len(procpad) < len(proc):
            raise ValueError("Padded processed signal is too short.")

        cleanpad[int(delay[0]) : int(len(clean) + int(delay[0])), 0] = clean[:, 0]
        cleanpad[int(delay[1]) : int(len(clean) + int(delay[1])), 1] = clean[:, 1]
        procpad[: len(proc)] = proc

        sii = mbstoi(
            cleanpad[:, 0],
            cleanpad[:, 1],
            procpad[:, 0],
            procpad[:, 1],
            cfg["mbstoi"]["fs"],
            cfg["mbstoi"]["gridcoarseness"],
        )
        csv_lines.append([f"{scene}_{listener}_{system}", sii])  # type: ignore

    with open(sii_file, "w") as csv_f:
        csv_writer = csv.writer(
            csv_f, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL
        )
        for line in csv_lines:
            csv_writer.writerow(line)


@hydra.main(config_path=".", config_name="config")
def run(cfg: DictConfig) -> None:
    logger.info("Prediction with MSGB + MBSTOI for train set")
    run_HL_processing(cfg, cfg.train_path)
    run_calculate_SI(cfg, cfg.train_path)

    logger.info("Prediction with MSGB + MBSTOI for train_indep set")
    run_HL_processing(cfg, cfg.train_indep_path)
    run_calculate_SI(cfg, cfg.train_indep_path)

    logger.info("Prediction with MSGB + MBSTOI for test set")
    run_HL_processing(cfg, cfg.test_path)
    run_calculate_SI(cfg, cfg.test_path)

    logger.info("Prediction with MSGB + MBSTOI for test_indep set")
    run_HL_processing(cfg, cfg.test_indep_path)
    run_calculate_SI(cfg, cfg.test_indep_path)


if __name__ == "__main__":
    run()



import copy
import logging
import re

import numpy as np
from scipy.interpolate import interp1d

from clarity.enhancer.gha.gainrule_camfit import gainrule_camfit_compr


def get_gaintable(
    audiogram, noisegatelevels, noisegateslope, cr_level, max_output_level
):
    """Compute a gaintable for a given audiogram.

    Replaces MATLAB GUI interface of original OpenMHA software for
    gaintable_camfit_compr table calculation. Assumes two channels and
    that audiogram frequencies are identical at two ears.

    Args:
        audiogram (Audiogram): the audiogram for which to compute the gain table
        audf (list): audiogram frequencies for fitting
        noisegatelevels (ndarray): compression threshold levels for each frequency band
        noisegateslope (int): determines slope of gains below compression threshold
        cr_level (int): overall input level in dB for calculation of compression ratios
        max_output_level (int): maximum output level in dB

    Returns:
        dict: dim ndarray of gains

    """
    # Initialise parameters
    num_channels = 2  # only configured for two channels
    sFitmodel = {}

    # Fixed centre frequencies for amplification bands
    sFitmodel["frequencies"] = [
        177,
        297,
        500,
        841,
        1414,
        2378,
        4.0000e03,
        6.7270e03,
        11314,
    ]
    sFitmodel["edge_frequencies"] = [
        1.0000e-08,
        229.2793,
        385.3570,
        648.4597,
        1.0905e03,
        1.8337e03,
        3.0842e03,
        5.1873e03,
        8.7241e03,
        10000001,
    ]

    # Input levels in SPL for which to compute the gains
    sFitmodel["levels"] = np.arange(-10, 110 + 1)
    sFitmodel["channels"] = num_channels
    sFitmodel["side"] = "lr"

    # Calculate gains and compression ratios
    output = gainrule_camfit_compr(
        audiogram,
        sFitmodel,
        noisegatelevels,
        noisegateslope,
        cr_level,
        max_output_level,
    )

    sGt = {}
    sGt["sGt_uncorr"] = copy.deepcopy(output["sGt"])  # sGt without noisegate correction

    output = multifit_apply_noisegate(
        output["sGt"],
        sFitmodel["frequencies"],
        sFitmodel["levels"],
        output["noisegatelevel"],
        output["noisegateslope"],
    )

    # Reshape sGt here to suit cfg file input
    sGt["sGt"] = np.transpose(np.reshape(output["sGt"], (121, 18), order="F"))

    sGt["noisegatelevel"] = output["noisegatelevel"]
    sGt["noisegateslope"] = output["noisegateslope"]
    sGt["frequencies"] = sFitmodel["frequencies"]
    sGt["levels"] = sFitmodel["levels"]
    sGt["channels"] = num_channels

    return sGt


def format_gaintable(gaintable, noisegate_corr=True):
    """
    Format gaintable for insertion into cfg file as long string.

    Args:
        gaintable (ndarray): The gaintable to format
        noisegate_corr (boolean, optional): apply noisegate correction or do not
            (default: True)

    Returns:
        str: gaintable formatted for insertion into OpenMHA
            cfg file

    """
    if noisegate_corr:
        sGt = gaintable["sGt"]
    else:
        logging.warning("Noise gate correction not being applied to gain table.")
        sGt = gaintable["sGt_uncorr"]

    # Do not apply gains to 9th and 18th row
    logging.info("No application of gains to 9th and 18th row of table.")
    sGt[[8, -1], :] = 0

    # Format for inclusion in cfg file
    v = "["
    for k in range(0, np.shape(sGt)[0]):
        v += f"{sGt[k, :]};"
    v += "]"
    v = v.replace("\n", "")
    v = re.sub(" +", " ", v)  # remove extra white spaces
    # v = re.sub(". ", " ", v)  # remove point after integers

    formatted_sGt = v

    return formatted_sGt


def multifit_apply_noisegate(
    sGt, sFit_model_frequencies, sFit_model_levels, noisegate_level, noisegate_slope
):
    """Apply noisegate.

    Based on OpenMHA subfunction of libmultifit.m.

    Args:
        sGt (ndarray): gain array
        sFit_model_frequencies (list): FFT filterbank frequencies
        sFit_model_levels (ndarray): levels at which to calculate gains
        noisegate_level (ndarray): chosen compression threshold
        noisegate_slope (ndarray): determines slope below compression threshold

    Returns:
        ndarray: Noise signal

    """

    for i in [0, 1]:
        for kf in range(len(sFit_model_frequencies)):
            gain_noisegate = interp1d(
                sFit_model_levels, sGt[:, kf, i], fill_value="extrapolate"
            )(noisegate_level[kf, i])
            idx = [
                i for i, x in enumerate(sFit_model_levels < noisegate_level[kf, i]) if x
            ]
            sGt[idx, kf, i] = (
                sFit_model_levels[idx] - noisegate_level[kf, i]
            ) * noisegate_slope[kf, i] + gain_noisegate

    output = {}
    output["sGt"] = {}
    output["sGt"] = sGt
    output["noisegatelevel"] = noisegate_level
    output["noisegateslope"] = noisegate_slope

    return output



import os
from pathlib import Path

import gdown

TARGET_DIR = "clarity_data/demo"
PACKAGE_NAME = "data.tgz"


def get_demo_data(metadata_url, target_dir):
    gdown.download(metadata_url, PACKAGE_NAME, quiet=False)
    p = Path(target_dir)
    if p.exists() is False:
        p.mkdir(parents=True, exist_ok=True)
    os.system(f"tar -xvzf {PACKAGE_NAME} --directory {target_dir}/")
    os.system(f"rm {PACKAGE_NAME}")


def get_metadata_demo(target_dir=TARGET_DIR):
    url = "https://drive.google.com/uc?export=download&id=14KGm2GaRgwlrZvtmMwWTYu7itaRVQV8f"
    get_demo_data(url, target_dir)


def get_targets_demo(target_dir=TARGET_DIR):
    url = "https://drive.google.com/uc?export=download&id=1uu2Hes1fzqNaZSCiFNhxZM3bE_fAVKsD"
    get_demo_data(url, target_dir)


def get_interferers_demo(target_dir=TARGET_DIR):
    url = "https://drive.google.com/uc?export=download&id=1_ssD238Qv-EETzC0hJze7JhLE7bHyqwG"
    get_demo_data(url, target_dir)


def get_rooms_demo(target_dir=TARGET_DIR):
    url = "https://drive.google.com/uc?export=download&id=1FBC8DI4Ru-g3Set0fDzoKmXTqHqNXV8n"
    get_demo_data(url, target_dir)


def get_scenes_demo(target_dir=TARGET_DIR):
    url = "https://drive.google.com/uc?export=download&id=1PB0CfGXhpkYNk8HbE5lTWowm2016x6Hl"
    get_demo_data(url, target_dir)


def get_hrirs_demo(target_dir=TARGET_DIR):
    url = "https://drive.google.com/uc?export=download&id=1USrHLFhOE_jdAQcEKqumG3M5dEZVcwSd"
    get_demo_data(url, target_dir)



"""Temporal smearing component of MSBG model."""
import math

import numpy as np

FFT_SIZE = 512
FRAME_SIZE = 256
SHIFT = 64


def audfilt(rl, ru, sampfreq, asize=256):
    """Calculate an auditory filter array.

    Args:
        rl (float): broadening factor on the lower side
        ru (float): broadening factor on the upper side
        sampfreq (float): signal sampling frequency
        asize (int, optional): number of taps in filter (default: {256})

    Returns:
        ndarray: A bank of auditory filters stored as 2-d numpy array

    """
    asize = int(asize)
    aud_filter = np.zeros((asize, asize))

    aud_filter[0, 0] = 1.0
    # Dividing by the erb to remove spectral tilt from the excitation pattern
    aud_filter[0, 0] = aud_filter[0, 0] / ((rl + ru) / 2)

    g = np.zeros(asize)
    for i in np.linspace(1, asize - 1, asize - 1, dtype=int):
        f_hz = i * np.divide(sampfreq, (2 * asize))
        f_erb = 24.7 * ((f_hz * 0.00437) + 1.0)
        # For lower side of the filter
        jj = np.arange(0, i)
        g = np.abs((i - jj) / i) * (4.0 * f_hz / (f_erb * rl))
        aud_filter[i, jj] = (1 + g) * np.exp(-g)
        # For upper side of the filter and centre
        jj = np.arange(i, asize)
        g = abs((i - jj) / i) * (4.0 * f_hz / (f_erb * ru))
        aud_filter[i, jj] = (1 + g) * np.exp(-g)
        aud_filter[i, :] = np.divide(aud_filter[i, :], (f_erb * (rl + ru) / (2 * 24.7)))

    return aud_filter


def make_smear_mat3(rl, ru, fs):
    """Make the smearing filter matrix.

    Args:
        rl (float): filter broadening factor on the lower side
        ru (float): filter broadening factor on the upper side
        fs (float): sampling frequency


    Returns:
        ndarray: The FFT_SIZE/2 X FFT_SIZE/2 smearing filter matrix

    """
    # FFTSIZE is assumed to contain a factor of 4
    assert FFT_SIZE % 4 == 0
    nyquist = int(FFT_SIZE / 2)
    half_nyquist = int(FFT_SIZE / 4)

    f_normal = audfilt(1, 1, fs, nyquist)
    f_wide = audfilt(rl, ru, fs, nyquist)
    # Extend the normal matrix so that the left-divide works better
    f_next = np.concatenate((f_normal, np.zeros((nyquist, half_nyquist))), axis=1)

    for i in np.arange(half_nyquist, nyquist):
        f_next[i, nyquist : min(2 * i - 1, 3 * half_nyquist)] = f_normal[
            i, 2 * i - nyquist : max(1, 2 * i - 3 * half_nyquist) : -1
        ]

    # This is equivalent to multiplying (convolving) the inverse of the
    # normal filters with the wide filters.
    f_smear = np.real(np.linalg.lstsq(f_next, f_wide, rcond=-1)[0])

    # Pruning to remove the extra bit
    f_smear = f_smear[0:nyquist, :]

    return f_smear


def smear3(f_smear, inbuffer):
    """Direct translation of smear3.m from MSBG hearing loss model.

    Args:
        f_smear (ndarray): The FFT_SIZE/2 X FFT_SIZE/2 smearing filter matrix
        inbuffer (ndarray): signal with prepended tone and noise

    Returns:
        ndarray: outbuffer

    """
    inlength = len(inbuffer)
    inpointer = 0
    outbuffer = np.zeros(int(np.ceil(inlength / SHIFT) + 3) * SHIFT)
    outpointer = 0

    overlaps = int(FRAME_SIZE / SHIFT)
    outwave = np.zeros([FRAME_SIZE, overlaps])
    buffer = np.arange(0, overlaps)
    nyquist = int(FFT_SIZE / 2)

    window = 0.5 - 0.5 * np.cos(
        2 * np.pi * (np.arange(1, FRAME_SIZE + 1) - 0.5) / FRAME_SIZE
    )
    window = window / math.sqrt(1.5)

    samplecount = min(FRAME_SIZE, inlength - inpointer)
    inwave = inbuffer[inpointer + np.arange(0, samplecount)]
    while samplecount > 0:
        winwave = np.zeros(FFT_SIZE)
        winwave[0:FRAME_SIZE] = window * inwave.flatten()
        spectrum = np.fft.fft(winwave, FFT_SIZE)
        power = spectrum[0:nyquist] * np.conj(spectrum[0:nyquist])
        mag = np.sqrt(power)
        phasor = spectrum[0:nyquist] / (mag + (mag == 0))
        smeared = np.dot(f_smear, power)
        spectrum[0:nyquist] = np.sqrt(smeared) * phasor
        spectrum[nyquist] = 0
        spectrum[(nyquist + 1) : FFT_SIZE] = np.conj(spectrum[nyquist - 1 : 0 : -1])
        winwave = np.real(np.fft.ifft(spectrum, FFT_SIZE))
        outwave[:, buffer[0]] = winwave[0:FRAME_SIZE] * window
        outframe = np.zeros(SHIFT)
        j = 0
        for i in np.arange(0, overlaps):
            outframe += outwave[j + np.arange(0, SHIFT), buffer[i]]
            j += SHIFT

        outbuffer[outpointer + np.arange(0, SHIFT)] = outframe
        outpointer += SHIFT

        buffer = np.roll(buffer, 1)
        inwave = np.append(inwave[SHIFT:FRAME_SIZE], np.zeros(SHIFT))
        inpointer += SHIFT
        samplecount = min(SHIFT, inlength - inpointer)
        inframe = inbuffer[inpointer + np.arange(0, samplecount)]
        inwave[FRAME_SIZE - SHIFT : FRAME_SIZE - SHIFT + samplecount] = inframe

    for k in np.arange(overlaps - 1, 0, -1):
        outframe = np.zeros(SHIFT)
        j = (overlaps - 1 - k) * SHIFT
        for i in np.arange(0, k):
            outframe += outwave[j + np.arange(0, SHIFT), buffer[i]]
            j += SHIFT
        outbuffer[outpointer + np.arange(0, SHIFT)] = outframe
        outpointer += SHIFT

    return outbuffer


class Smearer:
    """Class to hold the re-usable smearing filter."""

    rl: np.ndarray
    ru: np.ndarray
    fs: np.ndarray
    f_smear: np.ndarray

    def __init__(self, rl, ru, fs):
        self.rl = rl
        self.ru = ru
        self.fs = fs
        self.f_smear = make_smear_mat3(rl, ru, fs)

    def smear(self, input_signal: np.ndarray) -> np.ndarray:
        """Smear a given input signal."""
        return smear3(self.f_smear, input_signal)



from dataclasses import dataclass

import numpy as np


@dataclass
class Audiogram:
    """Dataclass to represent an audiogram."""

    levels_l: np.ndarray
    levels_r: np.ndarray
    cfs: np.ndarray

    @property
    def severity(self):
        """Categorise HL severity level for the audiogram.

        Note that this categorisation is different from that of the British Society of
        Audiology, which recommends descriptors mild, moderate, severe
        and profound for average hearing threshold levels at 250, 500,
        1000, 2000 and 4000 Hz of 21-40 dB HL, 41-70 dB HL, 71-95 dB HL
        and > 95 dB HL, respectively (BSA Pure-tone air-conduction and
        bone-conduction threshold audiometry with and without masking
        2018).

        Returns:
            str: severity level, one of SEVERE, MODERATE, MILD, NOTHING
                for each ear [left, right]

        """
        # calculate mean hearing loss between 2 & 8 kHz
        impairment_freqs = np.logical_and(2000 <= self.cfs, self.cfs <= 8000)
        severity_levels = [None, None]
        for i, levels in enumerate([self.levels_l, self.levels_r]):
            impairment_degree = np.mean(levels[impairment_freqs])
            if impairment_degree > 56:
                severity_level = "SEVERE"
            elif impairment_degree > 35:
                severity_level = "MODERATE"
            elif impairment_degree > 15:
                severity_level = "MILD"
            else:
                severity_level = "NOTHING"
            severity_levels[i] = severity_level
        return severity_levels

    def select_subset_of_cfs(self, selected_cfs):
        """Make a new audiogram using a given subset of centre freqs.

        Note, any selected_cfs that do not exist in the original audiogram
        are simply ignored

        Args:
            selected_cfs (list): List of centre frequencies to include

        Returns:
            Audiogram: New audiogram with reduced set of frequencies

        """
        indices = [i for i, cf in enumerate(self.cfs) if cf in selected_cfs]
        return Audiogram(
            levels_l=self.levels_l[indices],
            levels_r=self.levels_r[indices],
            cfs=self.cfs[indices],
        )



import json
import os

import hydra
import numpy as np
from omegaconf import DictConfig
from tqdm import tqdm

from clarity.enhancer.gha.audiogram import Audiogram
from clarity.enhancer.gha.gha_interface import GHAHearingAid


@hydra.main(config_path=".", config_name="config")
def enhance(cfg: DictConfig) -> None:
    enhanced_folder = os.path.join(cfg["path"]["exp_folder"], "enhanced_signals")
    os.makedirs(enhanced_folder, exist_ok=True)
    scenes_listeners = json.load(open(cfg["path"]["scenes_listeners_file"]))
    listener_audiograms = json.load(open(cfg["path"]["listeners_file"]))

    enhancer = GHAHearingAid(**cfg["GHAHearingAid"])

    for scene in tqdm(scenes_listeners):
        for listener in scenes_listeners[scene]:
            # retrieve audiograms
            cfs = np.array(listener_audiograms[listener]["audiogram_cfs"])
            audiogram_left = np.array(
                listener_audiograms[listener]["audiogram_levels_l"]
            )
            audiogram_right = np.array(
                listener_audiograms[listener]["audiogram_levels_r"]
            )
            audiogram = Audiogram(
                cfs=cfs, levels_l=audiogram_left, levels_r=audiogram_right
            )

            enhancer.process_files(
                infile_names=[
                    f"{cfg['path']['scenes_folder']}/{scene}_mixed_CH{ch}.wav"
                    for ch in range(1, cfg["num_channels"] + 1)
                ],
                outfile_name=f"{enhanced_folder}/{scene}_{listener}_HA-output.wav",
                audiogram=audiogram,
                listener=listener,
            )


if __name__ == "__main__":
    enhance()



"""Clarity ambisonic scene rendering."""
import json
import logging
import math
import warnings
from pathlib import Path

import librosa
import numpy as np

# from clarity_core import signal as ccs
# from clarity_core.signal import SPEECH_FILTER
from scipy.io import loadmat, wavfile
from scipy.signal import convolve
from tqdm import tqdm

import clarity.data.HOA_tools_cec2 as hoa
from clarity.data.HOA_tools_cec2 import HOARotator
from clarity.data.utils import SPEECH_FILTER, better_ear_speechweighted_snr

logger = logging.getLogger(__name__)

SPEED_SOUND = 344  # in m/s at 21 degrees C
FS = 44100


def two_point_rotation(rotation: dict, origin: np.ndarray, duration: int) -> np.ndarray:
    """Perform rotation defined by two control points.

    Args:
        rotation (dict): rotation object from scene definition
        origin (ndarray): origin view vector
        duration (int): total number of samples to generate for

    Returns:
        np.ndarray: sequence of theta values per sample
    """
    angle_origin = np.arctan2(origin[1], origin[0])
    angles = [math.radians(r["angle"]) - angle_origin for r in rotation]
    logger.info("angles=%s", angles)
    theta = hoa.rotation_vector(
        angles[0], angles[1], duration, rotation[0]["sample"], rotation[1]["sample"]
    )
    return theta


def pad_signal_start_end(signal: np.ndarray, delay: int, duration: int) -> np.ndarray:
    """Pad signal at start and end.

    Args:
        signal (array-like): ambisonic signals
        delay (int): number of zeros to pad at start
        duration (int): desired duration after start and end padding

    Returns:
        array-like: padded signals
    """
    time_after_signal = duration - delay - signal.shape[0]
    if time_after_signal < 0:
        # Signal is too long, so we need to truncate it
        signal = signal[:time_after_signal, :]
        time_after_signal = 0

    signal_pad_front = np.zeros([delay, signal.shape[1]])
    signal_pad_back = np.zeros([time_after_signal, signal.shape[1]])
    padded_signal = np.concatenate([signal_pad_front, signal, signal_pad_back], axis=0)
    return padded_signal


class SceneRenderer:
    """Ambisonic scene rendering class.

    Contains methods for generating signals from pseudorandom datasets for CEC2
    """

    def __init__(
        self,
        paths,
        metadata,
        ambisonic_order,
        equalise_loudness,
        reference_channel,
        channel_norms,
    ):
        """Initialise SceneRenderer.

        Args:
            paths ():
            metadata ():
            ambisonic_order ():
            equalise_loudness ():
            reference_channel ():
            channel_norms ():
        """
        self.paths = paths
        self.metadata = metadata
        self.ambisonic_order = ambisonic_order
        self.equalise_loudness = equalise_loudness
        self.reference_channel = reference_channel
        self.channel_norms = channel_norms

        # Build the HOA rotator object with precomputed rotation matrices
        self.hoa_rotator = HOARotator(self.ambisonic_order, resolution=0.1)

        # Build dictionary for looking up room data
        with open(f"{self.metadata.room_definitions}", "r", encoding="utf-8") as f:
            rooms = json.load(f)
        self.room_dict = {room["name"]: room for room in rooms}

        # Fixed hrir metadata for hrir sets being used
        with open(self.metadata.hrir_metadata, encoding="utf-8") as f:
            self.metadata.hrir_metadata = json.load(f)

    def make_interferer_filename(self, interferer: dict, dataset) -> str:
        """Construct filename for an interferer.

        Args:
            interferer (dict):

        Returns:
            str: Filename for an interferer.
        """
        data_type = interferer["type"]
        stem = self.paths.interferers.format(dataset=dataset, type=data_type)
        return f"{stem}/{interferer['name']}"

    def prepare_interferer_paths(self, scene):
        """Make list of full path filenames for interferers in scene.

        Args:
            scene ():

        Returns:
            list: List of full path filenames for interferers in scene.
        """
        logger.info("number of interferers: %s", len(scene["interferers"]))
        interferer_filenames = [
            self.make_interferer_filename(interferer, scene["dataset"])
            for interferer in scene["interferers"]
        ]
        return interferer_filenames

    def load_interferer_hoairs(self, scene):
        """Loads and returns the interferer hoa irs for given scene.

        Args:
            scene ():

        Returns:
            list: List of inferior hoa irs for the given scene.
        """
        room_id = scene["room"]
        n_interferers = len(scene["interferers"])

        hoair_path = self.paths.hoairs.format(dataset=scene["dataset"])
        interferer_hoairs = [
            f"{hoair_path}/HOA_{room_id}_i{n}.wav" for n in range(1, n_interferers + 1)
        ]
        irs = [wavfile.read(filename)[1] for filename in interferer_hoairs]
        return irs

    def load_interferer_signals(self, scene):
        """Loads and returns interferer signals for given scene.

        Args:
            scene ():

        Returns:
            list: List of signals.
        """
        interferer_audio_paths = self.prepare_interferer_paths(scene)
        # NOTE: all interferer signals are assumed to by at the Clarity fs = 44100 Hz
        with warnings.catch_warnings():
            # Suppress annoying warning generated by librosa when reading mp3 files
            warnings.simplefilter("ignore")
            signals = [
                librosa.load(signal_path, sr=None)[0]
                for signal_path in interferer_audio_paths
            ]
        signal_starts = [interferer["offset"] for interferer in scene["interferers"]]
        signal_lengths = [
            interferer["time_end"] - interferer["time_start"]
            for interferer in scene["interferers"]
        ]

        signals = [
            signal[ss : (ss + sl)]
            for (signal, ss, sl) in zip(signals, signal_starts, signal_lengths)
        ]

        return signals

    def make_hoa_target_anechoic(self, target, room):
        """Make the HOA anechoic target.

        Applies an anechoic HOA IR that models a source straight in front of the listener.
        The signal is delayed to match the propagation delay of the room.

        Args:
            target ():
            room (dict):
        """
        # TODO: The delay does not appear to correctly align the signals as expected
        t_pos = np.array(room["target"]["position"])
        l_pos = np.array(room["listener"]["position"])
        distance = np.linalg.norm(t_pos - l_pos)
        samples_delay = int(distance / SPEED_SOUND * FS)
        logger.info("%s, %s", room["target"]["position"], room["listener"]["position"])
        logger.info("target signal delay = %s samples %s m)", samples_delay, distance)

        # Values below for a source directly ahead of the listener
        anechoic_ir = np.array(
            [
                [1.0, 0.0, 0.0, 1.7320509, 0.0, 0.0, -1.1180342]
                + [0.0, 1.9364915, 0.0, 0.0, 0.0, 0.0, -1.6201853]
                + [0.0, 2.09165, 0.0, 0.0, 0.0, 0.0, 1.125]
                + [0.0, -1.6770511, 0.0, 2.2185302, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                + [
                    1.6056539,
                    0.0,
                    -1.7343045,
                    0.0,
                    2.326814,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ]
                + [-1.1267347, 0.0, 1.6327935, 0.0, -1.7886358, 0.0, 2.4209614, 0.0]
                + [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                + [0.0, -1.601086, 0.0, 1.6638975, 0.0, -1.839508, 0.0, 20.63472]
            ]
        )
        n_chans = (self.ambisonic_order + 1) ** 2
        anechoic_ir = anechoic_ir[:n_chans]
        hoa_target_anechoic = hoa.ambisonic_convolve(
            target, anechoic_ir, self.ambisonic_order
        )
        # prepend with zeros to match the propagation delay
        hoa_target_anechoic = np.vstack(
            (np.zeros((samples_delay, n_chans)), hoa_target_anechoic)
        )
        logger.info(
            "%s, %s, %s", anechoic_ir.shape, target.shape, hoa_target_anechoic.shape
        )

        return hoa_target_anechoic

    def generate_hoa_signals(self, scene: dict) -> tuple:
        """Generates HOA signals.

        Args:
            scene (dict): scene definitions
        """
        room_id = scene["room"]
        room = self.room_dict[room_id]

        # Make HOA target signal
        # NOTE: Target signals assumed to be at the Clarity fs = 44100 Hz
        target_audio_path = self.paths.targets.format(dataset=scene["dataset"])
        target_audio_path = f"{target_audio_path}/{scene['target']['name']}.wav"
        target, _fs = librosa.load(target_audio_path, sr=None)

        # TODO: set target to a fixed reference level??
        target_filt = convolve(target, SPEECH_FILTER, mode="full", method="fft")
        # rms of the target after speech weighted filter
        target_rms = np.sqrt(np.mean(target_filt**2))
        logger.info("target rms: %s", target_rms)
        target_hoair = self.paths.hoairs.format(dataset=scene["dataset"])
        target_hoair = f"{target_hoair}/HOA_{room_id}_t.wav"
        _fs, ir = wavfile.read(target_hoair)

        hoa_target = hoa.ambisonic_convolve(target, ir, self.ambisonic_order)

        hoa_target = pad_signal_start_end(
            hoa_target, scene["target"]["time_start"], scene["duration"]
        )

        # Make anechoic HOA target signal
        hoa_target_anechoic = self.make_hoa_target_anechoic(target, room)

        hoa_target_anechoic = pad_signal_start_end(
            hoa_target_anechoic, scene["target"]["time_start"], scene["duration"]
        )

        # Make HOA interferer signals
        irs = self.load_interferer_hoairs(scene)
        interferers = self.load_interferer_signals(scene)

        hoa_interferers = [
            hoa.ambisonic_convolve(signal, ir, order=self.ambisonic_order)
            for (signal, ir) in zip(interferers, irs)
        ]

        padded_interferers = [
            pad_signal_start_end(signal, interferer["time_start"], scene["duration"])
            for signal, interferer in zip(hoa_interferers, scene["interferers"])
        ]

        if self.equalise_loudness:
            padded_interferers = hoa.equalise_rms_levels(padded_interferers)

        flat_hoa_interferers = sum(padded_interferers)

        logger.info(
            "hoa_target.shape=%s; flat_hoa_interferers.shape=%s",
            hoa_target.shape,
            # flat_hoa_interferers.shape,
            flat_hoa_interferers,
        )

        th = two_point_rotation(
            scene["listener"]["rotation"],
            room["listener"]["view_vector"],
            scene["duration"],
        )
        hoa_target_anechoic_rotated = self.hoa_rotator.rotate(hoa_target_anechoic, th)
        hoa_target_rotated = self.hoa_rotator.rotate(hoa_target, th)
        flat_hoa_interferers_rotated = self.hoa_rotator.rotate(flat_hoa_interferers, th)

        return (
            hoa_target_rotated,
            flat_hoa_interferers_rotated,
            hoa_target_anechoic_rotated,
            th,
        )

    def save_signal_16bit(
        self, filename: str, signal: np.ndarray, norm: float = 1.0
    ) -> None:
        """Saves a signal to a 16 bit wav file.

        Args:
            filename (string): filename
            signal (np.array): signal
            norm (float): normalisation factor
        """
        signal /= norm
        n_clipped = np.sum(np.abs(signal) > 1.0)
        if n_clipped > 0:
            logger.warning("Writing %s: %s samples clipped", filename, n_clipped)
            np.clip(signal, -1.0, 1.0, out=signal)
        signal_16 = (32767 * signal).astype(np.int16)
        wavfile.write(filename, FS, signal_16)

    def generate_binaural_signals(
        self,
        scene: dict,
        hoa_target: np.ndarray,
        hoa_interferer: np.ndarray,
        hoa_target_anechoic: np.ndarray,
        out_path: str,
    ) -> None:
        """Generate and write binaural signals.

        Args:
            scene (dict): scene definitions
            hoa_target (ndarray): target signal in HOA domain
            hoa_interferer (ndarray): interferer signal in HOA domain
            hoa_target_anechoic (ndarray): anechoic target signal in HOA domain
            out_path (string): output path
        """
        ref_chan = self.reference_channel  # channel at which SNR is computed

        # Load all hrirs - one for each microphone pair
        hrir_filenames = [
            f"{self.paths.hrirs}/{name}.mat"
            for name in scene["listener"]["hrir_filename"]
        ]
        hrirs = [loadmat(hrir_filename) for hrir_filename in hrir_filenames]

        # Target and (flattened) interferer mixed down to binaural using each set of hrirs
        targets = [
            hoa.binaural_mixdown(hoa_target, hrir, self.metadata.hrir_metadata)
            for hrir in hrirs
        ]
        interferers = [
            hoa.binaural_mixdown(hoa_interferer, hrir, self.metadata.hrir_metadata)
            for hrir in hrirs
        ]

        target_anechoic = hoa.binaural_mixdown(
            hoa_target_anechoic,
            hrirs[ref_chan],  # Uses the reference channel's HRIR
            self.metadata.hrir_metadata,
        )

        # Measure pre-scaled SNR at reference channel
        start_time = scene["target"]["time_start"]
        end_time = scene["target"]["time_end"]
        sw_snr = better_ear_speechweighted_snr(
            targets[ref_chan][start_time:end_time, :],
            interferers[ref_chan][start_time:end_time, :],
        )

        # Scale interferers to achieve desired SNR at reference channel
        desired_snr = scene["SNR"]
        for interferer in interferers:
            interferer *= sw_snr * hoa.dB_to_gain(-desired_snr)

        # Make mixture by summing target and interferers
        mix = [t + i for t, i in zip(targets, interferers)]
        norm = np.max(mix)

        # Save all signal types for all channels
        norms = self.channel_norms
        out_path = out_path.format(dataset=scene["dataset"])
        file_stem = f"{out_path}/{scene['scene']}"
        for channel, (t, i, m, norm) in enumerate(
            zip(targets, interferers, mix, norms)
        ):
            for sig, sig_type in zip([t, i, m], ["target", "interferer", "mix"]):
                self.save_signal_16bit(
                    f"{file_stem}_{sig_type}_CH{channel}.wav", sig, norm
                )

        # Save the anechoic reference signal. Level normalised to abs max 1.0
        norm = np.max(np.abs(target_anechoic))
        self.save_signal_16bit(
            f"{file_stem}_target_anechoic_CH{ref_chan}.wav", target_anechoic, norm
        )

    def render_scenes(self, scenes: dict):
        """Renders scenes.

        Args:
            scenes (dict): scene definitions
        """
        # Make all necessary output directories
        output_path = self.paths.scenes
        for dataset in set(s["dataset"] for s in scenes):
            full_output_path = Path(output_path.format(dataset=dataset))
            if not full_output_path.exists():
                full_output_path.mkdir(parents=True, exist_ok=True)

        for scene in tqdm(scenes):
            # Stage 1: Generate the rotated signals in the HOA domain
            target, interferers, anechoic, head_turn = self.generate_hoa_signals(scene)

            # Save the head rotation signal
            head_turn = (np.mod(head_turn, 2 * np.pi) - np.pi) / np.pi
            file_stem = (
                output_path.format(dataset=scene["dataset"]) + "/" + scene["scene"]
            )
            wavfile.write(f"{file_stem}_hr.wav", 44100, head_turn)

            # Stage 2: Mix down to the binaural domain
            self.generate_binaural_signals(
                scene, target, interferers, anechoic, output_path
            )



"""Compressor Class"""
from __future__ import annotations

from typing import Any, List, Tuple

import numpy as np


class Compressor:
    def __init__(
        self,
        fs: int = 44100,
        attack: int = 5,
        release: int = 20,
        threshold: int = 1,
        attenuation: float = 0.0001,
        rms_buffer_size: float = 0.2,
        makeup_gain: int = 1,
    ):
        """Instantiate the Compressor Class.

        Args:
            fs (int): (default = 44100)
            attack (int): (default = 5)
            release int: (default = 20)
            threshold (int): (default = 1)
            attenuation (float): (default = 0.0001)
            rms_buffer_size (float): (default = 0.2)
            makeup_gain (int): (default = 1)
        """
        self.fs = fs
        self.rms_buffer_size = rms_buffer_size
        self.set_attack(attack)
        self.set_release(release)
        self.threshold = threshold
        self.attenuation = attenuation
        self.eps = 1e-8
        self.makeup_gain = makeup_gain

        # window for computing rms
        self.win_len = int(self.rms_buffer_size * self.fs)
        self.window = np.ones(self.win_len)

    def set_attack(self, t_msec: float) -> None:
        """DESCRIPTION

        Args:
            t_msec (float): DESCRIPTION

        Returns:
            float: DESCRIPTION
        """
        t_sec = t_msec / 1000
        reciprocal_time = 1 / t_sec
        self.attack = reciprocal_time / self.fs

    def set_release(self, t_msec: float) -> None:
        """DESCRIPTION

        Args:
            t_msec (float): DESCRIPTION

        Returns:
            float: DESCRIPTION
        """
        t_sec = t_msec / 1000
        reciprocal_time = 1 / t_sec
        self.release = reciprocal_time / self.fs

    def process(self, signal: np.ndarray) -> Tuple[Any, np.ndarray, List[Any]]:
        """DESCRIPTION

        Args:
            signal (np.array): DESCRIPTION

        Returns:
            np.array: DESCRIPTION
        """
        padded_signal = np.concatenate((np.zeros(self.win_len - 1), signal))
        rms = np.sqrt(
            np.convolve(padded_signal**2, self.window, mode="valid") / self.win_len
            + self.eps
        )
        comp_ratios: list = []
        curr_comp: float = 1.0
        for rms_i in rms:
            if rms_i > self.threshold:
                temp_comp = (rms_i * self.attenuation) + (
                    (1 - self.attenuation) * self.threshold
                )
                curr_comp = (curr_comp * (1 - self.attack)) + (temp_comp * self.attack)
            else:
                curr_comp = (1 * self.release) + curr_comp * (1 - self.release)
            comp_ratios.append(curr_comp)
        return (signal * np.array(comp_ratios) * self.makeup_gain), rms, comp_ratios



import logging
import os
import pathlib
import subprocess
import tempfile

import numpy as np
import soundfile
from jinja2 import Environment, FileSystemLoader
from soundfile import SoundFile

from clarity.enhancer.gha.gha_utils import format_gaintable, get_gaintable


class GHAHearingAid:
    def __init__(
        self,
        fs=44100,
        ahr=20,
        audf=None,
        cfg_file="prerelease_combination4_smooth",
        noisegatelevels=None,
        noisegateslope=0,
        cr_level=0,
        max_output_level=100,
        equiv0dBSPL=100,
        test_nbits=16,
    ):

        if audf is None:
            audf = [250, 500, 1000, 2000, 3000, 4000, 6000, 8000]
        if noisegatelevels is None:
            noisegatelevels = [38, 38, 36, 37, 32, 26, 23, 22, 8]

        self.fs = fs
        self.ahr = ahr
        self.audf = audf
        self.cfg_file = cfg_file
        self.noisegatelevels = noisegatelevels
        self.noisegateslope = noisegateslope
        self.cr_level = cr_level
        self.max_output_level = max_output_level
        self.equiv0dBSPL = equiv0dBSPL
        self.test_nbits = test_nbits

    def create_configured_cfgfile(
        self, input_file, output_file, formatted_sGt, cfg_template_file
    ):
        """Using Jinja2, generates cfg file for given configuration.

        Creates template output file and configures with correct filenames, peak level
        out and DC gaintable.

        Args:
            input_file (str): file to process
            output_file (str): file in which to store processed file
            formatted_sGt (ndarray): gaintable formatted for input into cfg file
            cfg_template_file: configuration file template
            ahr (int): amplification headroom

        Returns:
            cfg_filename (str): cfg filename
        """

        if self.fs != 44100:
            logging.error("Current GHA configuration requires 44.1kHz sampling rate.")

        cfg_template_file = pathlib.Path(cfg_template_file)

        # Define cfg filenames
        # Read new file and replace any parameter values necessary
        # Update peaklevel out by adding headroom
        logging.info("Adding %s dB headroom", self.ahr)

        peaklevel_in = int(self.equiv0dBSPL)
        peaklevel_out = int(self.equiv0dBSPL + self.ahr)

        # Render jinja2 template
        file_loader = FileSystemLoader(cfg_template_file.parent)
        env = Environment(loader=file_loader)
        template = env.get_template(cfg_template_file.name)
        output = template.render(
            io_in=input_file,
            io_out=output_file,
            peaklevel_in=f"[{peaklevel_in} {peaklevel_in} {peaklevel_in} {peaklevel_in}]",
            peaklevel_out=f"[{peaklevel_out} {peaklevel_out}]",
            gtdata=formatted_sGt,
        )

        return output

    def process_files(self, infile_names, outfile_name, audiogram, listener=None):
        """Process a set of input signals and generate an output.

        Args:
            infile_names (list[str]): List of input wav files. One stereo wav
                file for each hearing device channel
            outfile_name (str): File in which to store output wav files
            dry_run (bool): perform dry run only
        """
        logging.info("Processing %s with listener %s", outfile_name, listener)

        logging.info("Audiogram severity is %s", audiogram.severity)
        audiogram = audiogram.select_subset_of_cfs(self.audf)

        # Get gain table with noisegate correction
        gaintable = get_gaintable(
            audiogram,
            self.noisegatelevels,
            self.noisegateslope,
            self.cr_level,
            self.max_output_level,
        )
        formatted_sGt = format_gaintable(gaintable, noisegate_corr=True)

        cfg_template = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "cfg_files",
            f"{self.cfg_file}_template.cfg",
        )

        # Merge CH1 and CH3 files. This is the baseline configuration.
        # CH2 is ignored.
        fd_merged, merged_filename = tempfile.mkstemp(
            prefix="clarity-merged-", suffix=".wav"
        )
        # Only need file name; must immediately close the unused file handle.
        os.close(fd_merged)

        self.create_HA_inputs(infile_names, merged_filename)

        # Create the openMHA config file from the template
        fd_cfg, cfg_filename = tempfile.mkstemp(
            prefix="clarity-openmha-", suffix=".cfg"
        )
        # Again, only need file name; must immediately close the unused file handle.
        os.close(fd_cfg)

        with open(cfg_filename, "w", encoding="utf-8") as f:
            f.write(
                self.create_configured_cfgfile(
                    merged_filename, outfile_name, formatted_sGt, cfg_template
                )
            )

        # Process file using configured cfg file
        # Suppressing OpenMHA output with -q - comment out when testing
        # Append log of OpenMHA commands to /cfg_files/logfile
        subprocess.run(
            [
                "mha",
                "-q",
                "--log=logfile.txt",
                f"?read:{cfg_filename}",
                "cmd=start",
                "cmd=stop",
                "cmd=quit",
            ],
            check=True,
        )

        # Delete temporary files.
        os.remove(merged_filename)
        os.remove(cfg_filename)

        # Check output signal has energy in every channel
        sig = self.read_signal(outfile_name)

        if len(np.shape(sig)) == 1:
            sig = np.expand_dims(sig, axis=1)

        if not np.all(np.sum(abs(sig), axis=0)):
            raise ValueError("Channel empty.")

        self.write_signal(outfile_name, sig, floating_point=True)

        logging.info("OpenMHA processing complete")

    def read_signal(
        self, filename, offset=0, nsamples=-1, nchannels=0, offset_is_samples=False
    ):
        """Read a wavefile and return as numpy array of floats.

        Args:
            filename (string): Name of file to read
            offset (int, optional): Offset in samples or seconds (from start). Defaults to 0.
            nchannels: expected number of channel (default: 0 = any number OK)
            offset_is_samples (bool): measurement units for offset (default: False)
        Returns:
            ndarray: audio signal
        """
        try:
            wave_file = SoundFile(filename)
        except Exception as e:
            # Ensure incorrect error (24 bit) is not generated
            raise Exception(f"Unable to read {filename}.") from e

        if nchannels not in (0, wave_file.channels):
            raise Exception(
                f"Wav file ({filename}) was expected to have {nchannels} channels."
            )

        if wave_file.samplerate != self.fs:
            raise Exception(f"Sampling rate is not {self.fs} for filename {filename}.")

        if not offset_is_samples:  # Default behaviour
            offset = int(offset * wave_file.samplerate)

        if offset != 0:
            wave_file.seek(offset)

        x = wave_file.read(frames=nsamples)

        return x

    def write_signal(self, filename, x, floating_point=True):
        """Write a signal as fixed or floating point wav file."""

        if floating_point is False:
            if self.test_nbits == 16:
                subtype = "PCM_16"
                # If signal is float and we want int16
                x *= 32768
                x = x.astype(np.dtype("int16"))
                assert np.max(x) <= 32767 and np.min(x) >= -32768
            elif self.test_nbits == 24:
                subtype = "PCM_24"
            else:
                raise ValueError("test_nbits must be 16 or 24")
        else:
            subtype = "FLOAT"

        soundfile.write(filename, x, self.fs, subtype=subtype)

    def create_HA_inputs(self, infile_names, merged_filename):
        """Create input signal for baseline hearing aids."""

        if (infile_names[0][-5] != "1") or (infile_names[2][-5] != "3"):
            raise Exception("HA-input signal error: channel mismatch!")

        signal_CH1 = self.read_signal(infile_names[0])
        signal_CH3 = self.read_signal(infile_names[2])

        merged_signal = np.zeros((len(signal_CH1), 4))
        # channel index 0 = front microphone on the left hearing aid
        merged_signal[:, 0] = signal_CH1[:, 0]
        # channel index 1 = front microphone on the right hearing aid
        merged_signal[:, 1] = signal_CH1[:, 1]
        # channel index 2 = rear microphone on the left hearing aid
        merged_signal[:, 2] = signal_CH3[:, 0]
        # channel index 3 = rear microphone on the right hearing aid
        merged_signal[:, 3] = signal_CH3[:, 1]

        self.write_signal(merged_filename, merged_signal, floating_point=True)



"""
An FIR-based torch implementation of approximated MSBG hearing loss model
"""
import json
import os

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchaudio
from scipy.fftpack import fft
from scipy.interpolate import interp1d
from scipy.signal import ellip, firwin, firwin2, freqz

from clarity.evaluator.msbg.msbg_utils import (
    DF_ED,
    FF_ED,
    HZ,
    ITU_ERP_DRP,
    ITU_HZ,
    MIDEAR,
)

EPS = 1e-8
# old msbg matlab
# set RMS so that peak of output file so that no clipping occurs, set so that
# equiv0dBfileSPL > 100dB for LOUD input files
ref_RMSdB = -31.2
calib_dBSPL = (
    65  # what RMS of INPUT speech file translates to in real world (unweighted)
)
equiv0dBfileSPL = (
    calib_dBSPL - ref_RMSdB
)  # what 0dB file signal would translate to in dB SPL: constant for cochlear_simulate function

# clarity msbg
equiv0dBSPL = 100
ahr = 20
equiv_0dB_SPL = equiv0dBSPL + ahr


def generate_key_percent(signal, thr_dB, winlen):
    if winlen != np.floor(winlen):
        winlen = np.int(np.floor(winlen))
        print("\nGenerate_key_percent: \t Window length must be integer")

    siglen = len(signal)
    expected = thr_dB.copy()  # expected threshold
    non_zero = 10.0 ** ((expected - 30) / 10)  # put floor into histogram distribution

    nframes = 0
    totframes = np.int(np.floor(siglen / winlen))
    every_dB = np.zeros(totframes)

    for ix in range(totframes):
        start = ix * winlen
        this_sum = np.sum(signal[start : start + winlen] ** 2)  # sum of squares
        every_dB[nframes] = 10 * np.log10(non_zero + this_sum / winlen)
        nframes += 1

    used_thr_dB = expected.copy()

    # histogram should produce a two-peaked curve: thresh should be set in valley
    # between the two peaks, and set a bit above that, as it heads for main peak
    frame_idx = np.where(every_dB >= expected)[0]
    valid_frames = len(frame_idx)
    key = np.zeros(valid_frames * winlen, dtype=np.int)

    # convert frame numbers into indices for signal
    for ix in range(valid_frames):
        key[ix * winlen : ix * winlen + winlen] = np.arange(
            frame_idx[ix] * winlen, frame_idx[ix] * winlen + winlen, dtype=np.int
        )
    return key, used_thr_dB


def measure_rms(signal, sr, dB_rel_rms):
    """Measures toatal power of all 10 msec frams that are above a user-specified threshold

    Args:
        signal: input signal
        sr: sampling rate
        dB_rel_rms: threshold relative to first-stage rms (if it is made of a 2*1 array, second value over rules.)
                    only single value supported currently
    Returns:
        tuple: The percentage of frames that are required to be tracked for measuring RMS (useful when DR compression
               changes histogram shape)
    """
    win_secs = 0.01
    # first RMS is of all signal
    first_stage_rms = np.sqrt(np.mean(signal**2))
    # use this RMS to generate key threshold to more accurate RMS
    key_thr_dB = np.max([20 * np.log10(first_stage_rms) + dB_rel_rms, -80])
    key, used_thr_dB = generate_key_percent(
        signal, key_thr_dB, np.int(np.round(win_secs * sr))
    )
    # active = 100.0 * len(key) / len(signal)
    rms = np.sqrt(np.mean(signal[key] ** 2))
    rel_dB_thresh = used_thr_dB - 20 * np.log10(rms)
    return rms, key, rel_dB_thresh


def makesmearmat3(rl, ru, sr):
    fftsize = 512
    nyquist = np.int(fftsize // 2)
    fnor = audfilt(1, 1, nyquist, sr)
    fwid = audfilt(rl, ru, nyquist, sr)
    fnext = np.hstack([fnor, np.zeros([nyquist, nyquist // 2])])

    for i in np.arange(nyquist // 2 + 1, nyquist + 1, dtype=np.int):
        fnext[i - 1, nyquist : np.min([2 * i - 1, 3 * nyquist // 2])] = np.flip(
            fnor[
                i - 1, np.max([1, 2 * i - 3 * nyquist // 2]) - 1 : (2 * i - nyquist - 1)
            ]
        )
    fsmear = np.linalg.lstsq(fnext, fwid)[
        0
    ]  # https://stackoverflow.com/questions/33559946/numpy-vs-mldivide-matlab-operator
    fsmear = fsmear[:nyquist, :]

    return fsmear


def audfilt(rl, ru, size, sr):
    """Calculate an auditory filter array

    Args:
        rl: broadening factor on the lower side
        ru: broadening factor on the upper side
        size:
        sr:

    Returns:
        np.ndarray
    """
    aud_filter = np.zeros([size, size])
    aud_filter[0, 0] = 1.0
    aud_filter[0, 0] = aud_filter[0, 0] / ((rl + ru) / 2)

    g = np.zeros(size)
    for i in np.arange(2, size + 1, 1, dtype=np.int):
        fhz = (i - 1) * sr / (2 * size)
        erbhz = 24.7 * ((fhz * 0.00437) + 1)
        pl = 4 * fhz / (erbhz * rl)
        pu = 4 * fhz / (erbhz * ru)
        j = np.arange(1, i, dtype=np.int)
        g[j - 1] = np.abs((i - j) / (i - 1))
        aud_filter[i - 1, j - 1] = (1 + (pl * g[j - 1])) * np.exp(-pl * g[j - 1])
        j = np.arange(i, size + 1, dtype=np.int)
        g[j - 1] = np.abs((i - j) / (i - 1))
        aud_filter[i - 1, j - 1] = (1 + (pu * g[j - 1])) * np.exp(-pu * g[j - 1])
        aud_filter[i - 1, :] = aud_filter[i - 1, :] / (erbhz * (rl + ru) / (2 * 24.7))
    return aud_filter


class MSBGHearingModel(nn.Module):
    def __init__(
        self,
        audiogram,
        audiometric,
        sr=44100,
        spl_cali=True,
        src_posn="ff",
        kernel_size=1025,
        device=None,
    ):
        super().__init__()
        self.sr = sr
        self.spl_cali = spl_cali
        self.src_posn = src_posn
        self.kernel_size = kernel_size
        if device is None:
            self.device = "cuda" if torch.cuda.is_available() else "cpu"
        else:
            self.device = device

        # settings for audiogram

        audiogram = np.array(audiogram)
        # audiometric = np.array([250, 500, 1000, 2000, 4000, 6000])
        audiometric = np.array(audiometric)
        audiogram = np.append(audiogram, audiogram[-1])
        audiometric = np.append(audiometric, 16000)
        audiogram = np.append(audiogram[0], audiogram)
        audiometric = np.append(125, audiometric)

        audiogram_cfs = (
            np.array([0.125, 0.25, 0.5, 0.75, 1, 1.5, 2, 3, 4, 6, 8, 10, 12, 14, 16])
            * 1000
        )
        interpf = interp1d(audiometric, audiogram)
        audiogram = interpf(audiogram_cfs)

        # settings for src_to_cochlea_filt

        if src_posn == "ff":
            src_corrn = FF_ED
        elif src_posn == "df":
            src_corrn = DF_ED
        elif src_posn == "ITU":
            interf_itu = interp1d(ITU_HZ, ITU_ERP_DRP)
            src_corrn = interf_itu(HZ)
        nyquist = sr / 2
        ixf_useful = np.where(HZ < nyquist)[0]
        hz_used = np.append(HZ[ixf_useful], nyquist)
        corrn = src_corrn - MIDEAR
        interf_corrn = interp1d(HZ, corrn)
        last_corrn = interf_corrn(nyquist)
        corrn_used = np.append(corrn[ixf_useful], last_corrn)
        corrn_forward = 10 ** (0.05 * corrn_used)
        corrn_backward = 10 ** (0.05 * -1 * corrn_used)
        n_wdw = np.int(2 * np.floor((sr / 16e3) * 368 / 2))
        coch_filter_forward = firwin2(
            n_wdw + 1, hz_used / nyquist, corrn_forward, window=("kaiser", 4)
        )
        coch_filter_backward = firwin2(
            n_wdw + 1, hz_used / nyquist, corrn_backward, window=("kaiser", 4)
        )
        self.coch_padding = len(coch_filter_forward) // 2
        self.coch_filter_forward = (
            torch.tensor(coch_filter_forward, dtype=torch.float32, device=self.device)
            .unsqueeze(0)
            .unsqueeze(1)
        )
        self.coch_filter_backward = (
            torch.tensor(coch_filter_backward, dtype=torch.float32, device=self.device)
            .unsqueeze(0)
            .unsqueeze(1)
        )

        # Settings for smearing

        catch_up = 105.0  # dBHL where impaired catches up with normal

        # recruitment simulation comes with 3 degrees of broadening of auditory filters:
        # different set of centre freqs between simulations.
        # check and categorise audiogram: currently ALWAYS recruit with x2 broadening:
        # it's the smearing that changes

        impaired_freqs = np.where((audiogram_cfs >= 2000) & (audiogram_cfs <= 8000))[0]
        impaired_degree = np.mean(audiogram[impaired_freqs])

        # impairment degree affects smearing simulation, and now recruitment,
        # (assuming we do not have too much SEVERE losses present)

        current_dir = os.path.dirname(os.path.abspath(__file__))
        gtf_dir = os.path.join(current_dir, "../evaluator/msbg/msbg_hparams")
        if impaired_degree > 56:
            severe_not_moderate = 1
            GT4Bankfile = os.path.join(
                gtf_dir, "GT4FBank_Brd3.0E_Spaced2.3E_44100Fs.json"
            )
            bw_broaden_coef = 3
        elif impaired_degree > 35:
            severe_not_moderate = 0
            GT4Bankfile = os.path.join(
                gtf_dir, "GT4FBank_Brd2.0E_Spaced1.5E_44100Fs.json"
            )
            bw_broaden_coef = 2
        elif impaired_degree > 15:
            severe_not_moderate = -1
            GT4Bankfile = os.path.join(
                gtf_dir, "GT4FBank_Brd1.5E_Spaced1.1E_44100Fs.json"
            )
            bw_broaden_coef = 1
        else:
            severe_not_moderate = -2
            GT4Bankfile = os.path.join(
                gtf_dir, "GT4FBank_Brd1.5E_Spaced1.1E_44100Fs.json"
            )
            bw_broaden_coef = 1
        # GT4Bank = loadmat(GT4Bankfile)
        with open(GT4Bankfile, "r", encoding="utf-8") as fp:
            GT4Bank = json.load(fp)

        if severe_not_moderate > 0:
            fsmear = makesmearmat3(4, 2, sr)
        elif severe_not_moderate == 0:
            fsmear = makesmearmat3(2.4, 1.6, sr)
        elif severe_not_moderate == -1:
            fsmear = makesmearmat3(1.6, 1.1, sr)
        elif severe_not_moderate == -2:
            fsmear = makesmearmat3(1.001, 1.001, sr)

        self.smear_nfft = 512
        self.smear_win_len = 256
        self.smear_hop_len = 64
        smear_window = (
            0.5
            - 0.5
            * np.cos(
                2
                * np.pi
                * (np.arange(1, self.smear_win_len + 1) - 0.5)
                / self.smear_win_len
            )
        ) / np.sqrt(1.5)
        self.smear_window = torch.tensor(
            smear_window, dtype=torch.float32, device=self.device
        )
        self.fsmear = torch.tensor(fsmear, dtype=torch.float32, device=self.device)

        """ settings for recruitment"""
        cf_expnsn = 0 * np.array(GT4Bank["GTn_CentFrq"])
        eq_loud_db = 0 * np.array(GT4Bank["GTn_CentFrq"])
        for ix_cfreq in range(len(GT4Bank["GTn_CentFrq"])):
            if GT4Bank["GTn_CentFrq"][ix_cfreq] < audiogram_cfs[0]:
                cf_expnsn[ix_cfreq] = catch_up / (catch_up - audiogram[0])
            elif GT4Bank["GTn_CentFrq"][ix_cfreq] > audiogram_cfs[-1]:
                cf_expnsn[ix_cfreq] = catch_up / (catch_up - audiogram[-1])
            else:
                interp_audiog = interp1d(audiogram_cfs, audiogram)
                audiog_cf = interp_audiog(GT4Bank["GTn_CentFrq"][ix_cfreq])
                cf_expnsn[ix_cfreq] = catch_up / (catch_up - audiog_cf)
            eq_loud_db[ix_cfreq] = catch_up

        self.nchans = GT4Bank["NChans"]
        self.gtn_denoms = torch.tensor(
            GT4Bank["GTn_denoms"], dtype=torch.float32, device=self.device
        )
        self.gtn_nums = torch.tensor(
            GT4Bank["GTn_nums"], dtype=torch.float32, device=self.device
        )
        self.hp_denoms = torch.tensor(
            GT4Bank["HP_denoms"], dtype=torch.float32, device=self.device
        )
        self.hp_nums = torch.tensor(
            GT4Bank["HP_nums"], dtype=torch.float32, device=self.device
        )
        self.ngamma = int(GT4Bank["NGAMMA"])
        self.gtn_delays = GT4Bank["GTnDelays"]
        self.start2polehp = GT4Bank["Start2PoleHP"]

        erbn_centfrq = GT4Bank["ERBn_CentFrq"]
        chan_lpfB = []
        chan_lpfA = []
        fir_lpf = []
        for ixch in range(self.nchans):
            fc_envlp = (30 / 40) * np.min([100, erbn_centfrq[ixch]])
            chan_lpfB_ch, chan_lpfA_ch = ellip(2, 0.25, 35, fc_envlp / (self.sr / 2))
            chan_lpfB.append(chan_lpfB_ch)
            chan_lpfA.append(chan_lpfA_ch)
            fir_lpf_ch = firwin(
                self.kernel_size, fc_envlp / (self.sr / 2), pass_zero="lowpass"
            ) / np.sqrt(
                2
            )  # sqrt(2) is for the consistency with IIR
            fir_lpf.append(fir_lpf_ch)
        self.chan_lpfB = torch.tensor(
            np.array(chan_lpfB), dtype=torch.float32, device=self.device
        )
        self.chan_lpfA = torch.tensor(
            np.array(chan_lpfA), dtype=torch.float32, device=self.device
        )
        self.fir_lpf = torch.tensor(
            np.array(fir_lpf), dtype=torch.float32, device=self.device
        ).unsqueeze(1)

        self.expnsn_m1 = torch.tensor(
            cf_expnsn - 1, dtype=torch.float32, device=self.device
        )
        # self.envlp_max = torch.tensor(10 ** (0.05 * (eq_loud_db - equiv0dBfileSPL)),
        # dtype=torch.float32, device=self.device)
        self.envlp_max = torch.tensor(
            10 ** (0.05 * (eq_loud_db - equiv_0dB_SPL)),
            dtype=torch.float32,
            device=self.device,
        )

        recombination_dB = GT4Bank["Recombination_dB"]
        self.recruitmnet_out_coef = torch.tensor(
            10 ** (-0.05 * recombination_dB), dtype=torch.float32, device=self.device
        )

        "settings for FIR Gammatone Filters"
        gt_cfreq = np.array(GT4Bank["GTn_CentFrq"])
        gt_bw = np.array(GT4Bank["ERBn_CentFrq"]) * 1.1019 * bw_broaden_coef

        self.padding = (self.kernel_size - 1) // 2
        n_lin = torch.linspace(
            0, self.kernel_size - 1, self.kernel_size, device=self.device
        )
        window_ = 0.54 - 0.46 * torch.cos(2 * np.pi * n_lin / self.kernel_size)
        n_ = (
            torch.arange(
                0, self.kernel_size, dtype=torch.float32, device=self.device
            ).view(1, -1)
            / self.sr
        )
        center_hz = (
            torch.tensor(
                gt_cfreq / self.sr, dtype=torch.float32, device=self.device
            ).view(-1, 1)
            * self.sr
        )
        f_times_t = torch.matmul(center_hz, n_)
        carrier = torch.cos(2 * np.pi * f_times_t)
        carrier_sin = torch.sin(2 * np.pi * f_times_t)
        band_hz = (
            torch.tensor(gt_bw / self.sr, dtype=torch.float32, device=self.device).view(
                -1, 1
            )
            * self.sr
        )
        b_times_t = torch.matmul(band_hz, n_)
        kernel = torch.pow(n_, 4 - 1) * torch.exp(-2 * np.pi * b_times_t)
        gammatone = kernel * carrier

        self.peaks = torch.argmax(gammatone, dim=1)  # for gammatone delay calibration

        gammatone_sin = kernel * carrier_sin
        filters = (gammatone * window_).view(self.nchans, 1, self.kernel_size)
        # To get the normalised amplitude
        filters = filters.squeeze(1).cpu().numpy()
        fr_max = np.zeros(self.nchans)
        for i in range(self.nchans):
            fr = np.abs(fft(filters[i]))
            fr_ = fr[: int(self.kernel_size / 2)]
            fr_max[i] = np.max(fr_)
        amp = torch.tensor(fr_max, dtype=torch.float32, device=self.device)
        gammatone = gammatone / amp.unsqueeze(1)
        gammatone_sin = gammatone_sin / amp.unsqueeze(1)
        self.gt_fir = (gammatone * window_).view(self.nchans, 1, self.kernel_size)
        self.gt_fir_sin = (gammatone_sin * window_).view(
            self.nchans, 1, self.kernel_size
        )

        "settings for spl calibration"
        win_sec = 0.01
        self.db_relative_rms = -12
        self.win_len = int(self.sr * win_sec)

    def measure_rms(self, wav):
        bs = wav.shape[0]
        ave_rms = torch.sqrt(torch.mean(wav**2, dim=1) + EPS)
        thr_db = 20 * torch.log10(ave_rms + EPS) + self.db_relative_rms

        num_frames = wav.shape[1] // self.win_len
        wav_reshaped = torch.reshape(
            wav[:, : num_frames * self.win_len], [bs, num_frames, self.win_len]
        )
        db_frames = 10 * torch.log10(torch.mean(wav_reshaped**2, dim=2) + EPS)

        key_frames = (
            torch.where(
                db_frames > thr_db.unsqueeze(1),
                torch.tensor(1, dtype=torch.float32, device=self.device),
                torch.tensor(0, dtype=torch.float32, device=self.device),
            )
            .unsqueeze(-1)
            .repeat([1, 1, self.win_len])
            .reshape([bs, num_frames * self.win_len])
        )
        key_rms = torch.sqrt(
            torch.sum((wav[:, : num_frames * self.win_len] * key_frames) ** 2, dim=1)
            / (torch.sum(key_frames, dim=1) + EPS)
            + EPS
        )
        return key_rms.unsqueeze(1)

    def calibrate_spl(self, x):
        if self.spl_cali:
            levelreFS = 10 * torch.log10(torch.mean(x**2, dim=1, keepdim=True) + EPS)
            leveldBSPL = equiv_0dB_SPL + levelreFS
            rms = self.measure_rms(x)
            change_dB = leveldBSPL - (equiv_0dB_SPL + 20 * torch.log10(rms + EPS))
            x = x * 10 ** (0.05 * change_dB)
        return x

    def src_to_cochlea_filt(self, x, coch_filter):
        return F.conv1d(x, coch_filter, padding=self.coch_padding)

    def smear(self, x):
        """Padding issue needs to be worked out"""
        length = x.shape[2]
        x = x.view(x.shape[0], x.shape[2])
        spec = torch.stft(
            x,
            n_fft=self.smear_nfft,
            hop_length=self.smear_hop_len,
            win_length=self.smear_win_len,
            window=self.smear_window,
        )
        power = torch.square(spec[:, : self.smear_nfft // 2, :, 0]) + torch.square(
            spec[:, : self.smear_nfft // 2, :, 1]
        )
        mag = torch.sqrt(power + EPS).unsqueeze(-1)
        phasor = spec[:, : self.smear_nfft // 2, :, :] / (mag + EPS)

        smeared_power = (
            torch.matmul(power.transpose(-1, -2), self.fsmear.transpose(0, 1))
            .transpose(-1, -2)
            .unsqueeze(-1)
            + EPS
        )
        smeared_power = torch.clamp(smeared_power, min=0)
        smeared_spec_nyquist = torch.sqrt(smeared_power + EPS) * phasor
        smeared_spec_mid = torch.zeros(
            [smeared_power.shape[0], 1, smeared_power.shape[2], 2],
            dtype=torch.float32,
            device=self.device,
        )
        smeared_spec = torch.cat([smeared_spec_nyquist, smeared_spec_mid], dim=1)
        smeared_wav = torch.istft(
            smeared_spec,
            n_fft=self.smear_nfft,
            hop_length=self.smear_hop_len,
            win_length=self.smear_win_len,
            window=self.smear_window,
            length=length,
        )
        return smeared_wav.unsqueeze(1)

    def recruitment(self, x):
        nsamps = x.shape[-1]
        ixhp = 0
        outputs = []
        for ixch in range(self.nchans):
            # Gammaton filtering
            pass_n = torchaudio.functional.lfilter(
                x, self.gtn_denoms[ixch, :], self.gtn_nums[ixch, :]
            )
            for _ixg in range(self.ngamma - 1):
                pass_n = torchaudio.functional.lfilter(
                    pass_n, self.gtn_denoms[ixch, :], self.gtn_nums[ixch, :]
                )
            dly = self.gtn_delays[ixch]

            pass_n_cali = torch.zeros_like(pass_n)
            pass_n_cali[:, :, : nsamps - dly] = pass_n[:, :, dly:nsamps]
            # Tail control
            if ixch >= self.start2polehp:
                ixhp += 1
                pass_n_cali = torchaudio.functional.lfilter(
                    pass_n_cali, self.hp_denoms[ixhp - 1, :], self.hp_nums[ixhp - 1, :]
                )

            # Get the envelope
            envlp_out = torchaudio.functional.lfilter(
                torch.abs(pass_n_cali), self.chan_lpfA[ixch, :], self.chan_lpfB[ixch, :]
            )
            envlp_out = torch.flip(envlp_out, dims=[-1])
            envlp_out = torchaudio.functional.lfilter(
                envlp_out, self.chan_lpfA[ixch, :], self.chan_lpfB[ixch, :]
            )
            envlp_out = torch.flip(envlp_out, dims=[-1])

            envlp_out = torch.clamp(envlp_out, min=EPS, max=self.envlp_max[ixch])
            gain = (envlp_out / self.envlp_max[ixch]) ** self.expnsn_m1[ixch]
            outputs.append(gain * pass_n_cali)

        y = torch.stack(outputs, dim=-1).sum(dim=-1)
        y = y * self.recruitmnet_out_coef
        return y

    def recruitment_fir(self, x):
        nsamps = x.shape[-1]
        x = x.repeat([1, self.nchans, 1])
        real = F.conv1d(
            x, self.gt_fir, bias=None, padding=self.padding, groups=self.nchans
        )
        imag = F.conv1d(
            x, self.gt_fir_sin, bias=None, padding=self.padding, groups=self.nchans
        )
        real_cali = torch.zeros_like(real)
        imag_cali = torch.zeros_like(imag)
        for i in range(self.nchans):
            real_cali[:, i, : nsamps - self.peaks[i]] = real[
                :, i, self.peaks[i] : nsamps
            ]
            imag_cali[:, i, : nsamps - self.peaks[i]] = imag[
                :, i, self.peaks[i] : nsamps
            ]

        env = torch.sqrt(real_cali * real_cali + imag_cali * imag_cali + EPS)
        env = F.conv1d(
            env, self.fir_lpf, bias=None, padding=self.padding, groups=self.nchans
        )

        env_max = self.envlp_max.unsqueeze(0).unsqueeze(-1).repeat([1, 1, nsamps])
        gain = torch.clamp(env / env_max, min=EPS, max=1)
        gain = gain ** self.expnsn_m1.unsqueeze(0).unsqueeze(-1).repeat([1, 1, nsamps])
        y = torch.sum(gain * real_cali, dim=1, keepdim=True)
        y = y * self.recruitmnet_out_coef
        return y

    def forward(self, x):
        x = self.calibrate_spl(x)
        x = x.unsqueeze(1)
        x = self.src_to_cochlea_filt(x, self.coch_filter_forward)
        x = self.smear(x)
        # x = self.recruitment(x)
        x = self.recruitment_fir(x)
        y = self.src_to_cochlea_filt(x, self.coch_filter_backward)
        return y.squeeze(1)


class torchloudnorm(nn.Module):
    def __init__(
        self,
        sr=44100,
        norm_lufs=-36,
        kernel_size=1025,
        block_size=0.4,
        overlap=0.75,
        gamma_a=-70,
    ):
        super().__init__()
        self.sr = sr
        self.norm_lufs = norm_lufs
        self.kernel_size = kernel_size
        self.padding = kernel_size // 2
        self.device = "cuda" if torch.cuda.is_available() else "cpu"

        # for frequency weighting filters - account for the acoustic respose
        # of the head and auditory system
        pyln_high_shelf_b = np.array([1.53090959, -2.65116903, 1.16916686])
        pyln_high_shelf_a = np.array([1.0, -1.66375011, 0.71265753])

        # fir high_shelf
        w_high_shelf, h_high_shelf = freqz(pyln_high_shelf_b, pyln_high_shelf_a, fs=sr)
        freq_high_shelf = np.append(w_high_shelf, sr / 2)
        gain_high_shelf = np.append(np.abs(h_high_shelf), np.abs(h_high_shelf)[-1])
        fir_high_shelf = firwin2(kernel_size, freq_high_shelf, gain_high_shelf, fs=sr)

        # fir high_pass
        fc_high_pass = 38.0
        fir_high_pass = firwin(kernel_size, fc_high_pass, pass_zero="highpass", fs=sr)

        self.high_shelf = (
            torch.tensor(fir_high_shelf, dtype=torch.float32, device=self.device)
            .unsqueeze(0)
            .unsqueeze(1)
        )
        self.high_pass = (
            torch.tensor(fir_high_pass, dtype=torch.float32, device=self.device)
            .unsqueeze(0)
            .unsqueeze(1)
        )

        "rms measurement"
        self.frame_size = int(block_size * sr)
        self.frame_shift = int(block_size * sr * (1 - overlap))
        self.unfold = torch.nn.Unfold(
            (1, self.frame_size), stride=(1, self.frame_shift)
        )
        self.gamma_a = gamma_a

    def apply_filter(self, x):
        x = F.conv1d(x, self.high_shelf, padding=self.padding)
        x = F.conv1d(x, self.high_pass, padding=self.padding)
        return x

    def integrated_loudness(self, x):
        x = self.apply_filter(x)
        x_unfold = self.unfold(x.unsqueeze(2))

        z = (
            torch.sum(x_unfold**2, dim=1) / self.frame_size
        )  # mean square for each frame
        el = -0.691 + 10 * torch.log10(z + EPS)

        idx_a = torch.where(el > self.gamma_a, 1, 0)
        z_ave_gated_a = torch.sum(z * idx_a, dim=1, keepdim=True) / (
            torch.sum(idx_a, dim=1, keepdim=True) + 1e-8
        )
        gamma_r = -0.691 + 10 * torch.log10(z_ave_gated_a + EPS) - 10

        idx_r = torch.where(el > gamma_r, 1, 0)
        idx_a_r = idx_a * idx_r
        z_ave_gated_a_r = torch.sum(z * idx_a_r, dim=1, keepdim=True) / (
            torch.sum(idx_a_r, dim=1, keepdim=True) + 1e-8
        )
        lufs = -0.691 + 10 * torch.log10(z_ave_gated_a_r + EPS)  # loudness
        return lufs

    def normalize_loudness(self, x, lufs):
        delta_loudness = self.norm_lufs - lufs
        gain = torch.pow(10, delta_loudness / 20)
        return gain * x

    def forward(self, x):
        loudness = self.integrated_loudness(x.unsqueeze(1))
        y = self.normalize_loudness(x, loudness)
        return y



import json
import logging

import hydra
from omegaconf import DictConfig

from clarity.data.scene_renderer_cec2 import SceneRenderer

logger = logging.getLogger(__name__)


def render_scenes(cfg):
    for dataset in cfg.scene_renderer:
        logger.info(f"Beginning scene generation for {dataset} set...")
        with open(cfg.scene_renderer[dataset].metadata.scene_definitions, "r") as f:
            scenes = json.load(f)

        starting_scene = (
            cfg.scene_renderer[dataset].chunk_size * cfg.render_starting_chunk
        )
        n_scenes = (
            cfg.scene_renderer[dataset].chunk_size * cfg.render_n_chunk_to_process
        )
        scenes = scenes[starting_scene : starting_scene + n_scenes]

        scene_renderer = SceneRenderer(
            cfg.scene_renderer[dataset].paths,
            cfg.scene_renderer[dataset].metadata,
            **cfg.render_params,
        )
        scene_renderer.render_scenes(scenes)


@hydra.main(config_path=".", config_name="config")
def run(cfg: DictConfig) -> None:
    logger.info("Rendering scenes")
    render_scenes(cfg)


if __name__ == "__main__":
    run()



"""Gammatone filterbank simulation of the Cochlea."""

import logging
from dataclasses import dataclass
from typing import Dict, Tuple, Union

import numpy as np
import scipy
from scipy import signal

from clarity.evaluator.msbg.audiogram import Audiogram
from clarity.evaluator.msbg.msbg_utils import read_gtf_file
from clarity.evaluator.msbg.smearing import Smearer

# TODO: Fix power overflow error when (expansion_ratios[ixch] - 1) < 0


@dataclass
class FilterBank:
    """Holds the numerators and demoninators of an IIR filter bank."""

    nums: np.ndarray
    denoms: np.ndarray


# Parameters for smearing and gammatone filtering according to degree of loss
HL_PARAMS: Dict[str, Dict[str, Union[str, tuple]]] = {
    "SEVERE": {
        "gtfbank_file": "GT4FBank_Brd3.0E_Spaced2.3E_44100Fs",
        "smear_params": (4, 2),  # asymmetric severe smearing
    },
    "MODERATE": {
        "gtfbank_file": "GT4FBank_Brd2.0E_Spaced1.5E_44100Fs",
        "smear_params": (2.4, 1.6),  # asymmetric moderate smearing
    },
    "MILD": {
        "gtfbank_file": "GT4FBank_Brd1.5E_Spaced1.1E_44100Fs",
        "smear_params": (1.6, 1.1),  # asymmetric mild smearing
    },
    "NOTHING": {
        "gtfbank_file": "GT4FBank_Brd1.5E_Spaced1.1E_44100Fs",
        "smear_params": (1.0, 1.0),  # No smearing
    },
}


def compute_recruitment_parameters(
    gtn_cf: np.ndarray, audiogram: Audiogram, catch_up: float
) -> Tuple[np.ndarray, np.ndarray]:
    """Compute parameters to be used in recruitment model.

    Computes expansion ratios for each gammatone filterbank channel and
    the equal loudness catch up level per channel (currently this is
    a fixed value per channel)

    Args:
        gtn_cf (ndarray): gammatone filterbank centre frequencies
        audiogram (Audiogram): the audiogram to employ
        catch_up (float): level in dB at which catches up with NH

    Returns:
        ndarray: expansion ratio for each gammatone filterbank channel

    """
    cf_expansion = np.zeros(gtn_cf.shape)  # expansion ratios

    for ix_cf in np.arange(0, gtn_cf.shape[0]):
        if gtn_cf[ix_cf] < audiogram.cfs[0]:
            # Extend audiogram, flat below lowest freq measured
            cf_expansion[ix_cf] = catch_up / (catch_up - audiogram.levels[0])
        elif gtn_cf[ix_cf] > audiogram.cfs[-1]:
            # Extend audiogram, flat above highest freq measured
            cf_expansion[ix_cf] = catch_up / (catch_up - audiogram.levels[-1])
        else:
            # In the sensible region
            cf_to_level_fn = scipy.interpolate.interp1d(
                audiogram.cfs, audiogram.levels, kind="linear"
            )
            # Assumes catch-up at catch_up dB (typ 100-105)
            cf_expansion[ix_cf] = catch_up / (catch_up - cf_to_level_fn(gtn_cf[ix_cf]))

    #  Loudness catch-up levels - currently same for all channels
    eq_loud_db_catch_up = catch_up * np.ones(gtn_cf.shape)

    return cf_expansion, eq_loud_db_catch_up


def gammatone_filterbank(
    x: np.ndarray,
    ngamma: int,
    gtn_filters: FilterBank,
    gtn_delays: np.ndarray,
    start2poleHP: int,
    hp_filters: FilterBank,
) -> np.ndarray:
    """Pass signal through gammatone filterbank.

    Args:
        x (ndarray): input signal
        ngamma (int): 4, gammatone order
        gtn_filters (FilterBank): set of gammatone filters
        gtn_delays (ndarray): gammatone filter delays
        start2poleHP (int): parameter value from gtfbank_file
        hp_filter (FilterBarker): set of highpass filters

    Returns:
        ndarray: cochleagram with gtn_denoms.shape[0] channels of len(x)

    """
    n_chans = gtn_filters.denoms.shape[0]
    n_samples = len(x)

    cochleagram = np.zeros((n_chans, n_samples))

    # Implement Nth order Gammatone filterbank
    for ixch in np.arange(0, n_chans):
        pass_n = x
        for __ in np.arange(0, ngamma):
            pass_n = signal.lfilter(
                gtn_filters.nums[ixch, :], gtn_filters.denoms[ixch, :], pass_n
            )

        dly = gtn_delays[ixch]
        # Time-advance signal to compensate for IIR shift, zero-pad remainder
        pass_n[np.arange(0, (n_samples - dly))] = pass_n[np.arange(dly, n_samples)]
        pass_n[np.arange(n_samples - dly, n_samples)] = 0

        # ...possibly control lf tails with second order highpass
        if ixch >= (start2poleHP - 1):
            ix_hp = ixch - start2poleHP + 1
            pass_n = signal.lfilter(
                hp_filters.nums[ix_hp, :], hp_filters.denoms[ix_hp, :], pass_n
            )
        cochleagram[ixch, :] = pass_n
    return cochleagram


def compute_envelope(
    coch_sig: np.ndarray, erbn_cf: np.ndarray, fs: float
) -> np.ndarray:
    """Obtain signal envelope.

    Envelope computed using full-wave rectification and low-pass filter

    Args:
        coch_sig (ndarray): input signal
        erbn_cf (ndarray): ERB centre frequencies
        fs (float): sampling frequency

    Returns:
        ndarray: signal envelope

    """

    envelope = np.zeros(coch_sig.shape)
    n_chans = coch_sig.shape[0]
    for ixch in np.arange(0, n_chans):
        # Extract envelope; channel envelope lpf is NOT fixed for all channels,
        # tracks slightly above ERBn
        # Don't need to worry about phase since using bidirectional filters
        # Put limit on bandwidth for very high freq channels, (30/40) == -3dB at fc,
        fc_envlp = (30 / 40) * min(100, erbn_cf[ixch])

        # Reduced rate of cut(4-pole), but improved time response of filter.
        chan_lpfB, chan_lpfA = signal.ellip(2, 0.25, 35, fc_envlp / (fs / 2))

        # Full-wave rectify (take absolute values of raw signal coch_sig) and
        # low-pass to get envelope
        padlen = 3 * (max(len(chan_lpfA), len(chan_lpfB))) - 1
        envelope[ixch, :] = signal.filtfilt(
            chan_lpfB, chan_lpfA, np.abs(coch_sig[ixch, :]), padlen=padlen
        )
    return envelope


def recruitment(
    coch_sig: np.ndarray,
    envelope: np.ndarray,
    SPL_equiv_0dB: float,
    expansion_ratios: np.ndarray,
    eq_loud_db: np.ndarray,
) -> np.ndarray:
    """Simulate loudness recruitment.

    Args:
        coch_sig (ndarray): input signal
        envelope (ndarray): signal envelope
        SPL_equiv_0dB (float): equivalent level in dB SPL of 0 dB Full Scale
        expansion_ratios (ndarray): expansion ratios for expanding channel
            signals
        eq_loud_db (ndarray): loudness catch-up level in dB

    Returns:
        ndarray: cochlear output signal

    """

    n_chans = coch_sig.shape[0]
    for ixch in np.arange(0, n_chans):
        # Limit max envelope, and hence maximum gain
        envlp_out = envelope[ixch, :]
        envlp_max = np.power(10, 0.05 * (eq_loud_db[ixch] - SPL_equiv_0dB))
        envlp_out[envlp_out > envlp_max] = envlp_max
        envlp_out[envlp_out < 1e-9] = 1e-9

        gain = (envlp_out / envlp_max) ** (expansion_ratios[ixch] - 1)
        coch_sig[ixch, :] = coch_sig[ixch, :] * gain

    return coch_sig


class Cochlea:
    """Simulate the cochlea.

    Includes simulation of effects of bandwidth broadening (smearing) and recruitment.
    Implements 3 different degrees of impairment which affect the degree of smearing.
    Recruitment currently always with x2 broadening.

    Degree of hearing impairment used to control the following filterbank variables:
    BROADEN, SPACING, NGAMMA, Fs, N_Chans,
    ERBn_CentFrq, GTn_CentFrq, GTnDelays,
    GTn_denoms, GTn_nums, Start2PoleHP, HP_FCorner, HP_denoms,
    HP_nums, HP_Delays,  Recombination_dB
    """

    def __init__(
        self, audiogram: Audiogram, catch_up_level: float = 105.0, fs: float = 44100.0
    ):
        """Cochlea constructor.

        Args:
            audiogram (Audiogram): Audiogram characterising hearing loss
            catch_up_level (float, optional): loudness catch-up level in dB (default: {105})
            fs (float, optional): sampling frequency

        """
        self.fs = fs
        self.audiogram = audiogram  # Audiogram to employ
        self.catch_up_level = catch_up_level  # Level in dB at which catches up with NH

        # Compute severity level and set parameters accordingly
        severity_level = self.audiogram.severity
        self.gtfbank_params = read_gtf_file(
            f"msbg_hparams/{HL_PARAMS[severity_level]['gtfbank_file']}.json"
        )
        self.cf_expansion, self.eq_loud_db_catch_up = compute_recruitment_parameters(
            self.gtfbank_params["GTn_CentFrq"], audiogram, catch_up_level
        )

        # Set-up the smearer
        self.smearer = None
        if severity_level != "NOTHING":
            smear_params = HL_PARAMS[severity_level]["smear_params"]
            self.smearer = Smearer(smear_params[0], smear_params[1], fs)

        logging.info("Severity level - %s", severity_level)

    def simulate(self, coch_sig: np.ndarray, equiv_0dB_file_SPL: float) -> np.ndarray:
        """Pass a signal through the cochlea.

        Args:
            coch_sig (ndarray): input signal
            equiv_0dB_file_SPL (float): equivalent level in dB SPL of 0 dB Full Scale

        Returns:
            ndarray: cochlear output signal

        """
        if self.smearer is not None:
            coch_sig = self.smearer.smear(coch_sig)

        # Filter with gammatone filterbank
        coch_sig_out = gammatone_filterbank(
            coch_sig,
            self.gtfbank_params["NGAMMA"],
            FilterBank(
                self.gtfbank_params["GTn_nums"], self.gtfbank_params["GTn_denoms"]
            ),
            self.gtfbank_params["GTnDelays"],
            self.gtfbank_params["Start2PoleHP"],
            FilterBank(
                self.gtfbank_params["HP_nums"], self.gtfbank_params["HP_denoms"]
            ),
        )

        # Simulate loudness recruitment
        envelope = compute_envelope(
            coch_sig_out, np.array(self.gtfbank_params["ERBn_CentFrq"]), self.fs
        )

        coch_sig_out = recruitment(
            coch_sig_out,
            envelope,
            equiv_0dB_file_SPL,
            self.cf_expansion,
            self.eq_loud_db_catch_up,
        )

        # Recombine channels to obtain output signal
        coch_sig_out = np.sum(coch_sig_out, axis=0) * np.power(
            10, -0.05 * self.gtfbank_params["Recombination_dB"]
        )

        return coch_sig_out



import logging
import os

import hydra
from omegaconf import DictConfig

from clarity.data.scene_builder_cec2 import RoomBuilder, SceneBuilder, set_random_seed

logger = logging.getLogger(__name__)


def build_rooms_from_rpf(cfg):
    rb = RoomBuilder()
    for dataset in cfg.room_datasets:
        room_file = os.path.join(cfg.path.metadata_dir, f"rooms.{dataset}.json")
        if not os.path.exists(room_file):
            rb.build_from_rpf(**cfg.room_datasets[dataset])
            rb.save_rooms(room_file)
        else:
            logger.info(f"rooms.{dataset}.json exists, skip")


def instantiate_scenes(cfg):
    rb = RoomBuilder()
    set_random_seed(cfg.random_seed)
    for dataset in cfg.scene_datasets:
        scene_file = os.path.join(cfg.path.metadata_dir, f"scenes.{dataset}.json")
        if not os.path.exists(scene_file):
            logger.info(f"instantiate scenes for {dataset} set")
            room_file = os.path.join(cfg.path.metadata_dir, f"rooms.{dataset}.json")
            rb.load(room_file)
            sb = SceneBuilder(
                rb=rb,
                scene_datasets=cfg.scene_datasets[dataset],
                target=cfg.target,
                interferer=cfg.interferer,
                snr_range=cfg.snr_range[dataset],
                listener=cfg.listener,
                shuffle_rooms=cfg.shuffle_rooms,
            )
            sb.instantiate_scenes(dataset=dataset)
            sb.save_scenes(scene_file)
        else:
            logger.info(f"scenes.{dataset}.json exists, skip")


@hydra.main(config_path=".", config_name="config")
def run(cfg: DictConfig) -> None:
    logger.info("Building rooms")
    build_rooms_from_rpf(cfg)
    logger.info("Instantiating scenes")
    instantiate_scenes(cfg)


if __name__ == "__main__":
    run()



"""
Adopted from Asteroid
https://github.com/asteroid-team/asteroid/blob/master/asteroid/engine/system.py
"""
import pytorch_lightning as pl
from torch.optim.lr_scheduler import ReduceLROnPlateau


class System(pl.LightningModule):
    def __init__(
        self,
        model,
        optimizer,
        loss_func,
        train_loader,
        val_loader=None,
        scheduler=None,
        config=None,
    ):
        super().__init__()
        self.model = model
        self.optimizer = optimizer
        self.loss_func = loss_func
        self.train_loader = train_loader
        self.val_loader = val_loader
        self.scheduler = scheduler
        self.config = {} if config is None else config

    def forward(self, *args, **kwargs):
        """Applies forward pass of the model.

        Returns:
            :class:`torch.Tensor`
        """
        return self.model(*args, **kwargs)

    def common_step(self, batch, batch_nb, train=True):  # pylint: disable=W0613
        """Common forward step between training and validation.
        The function of this method is to unpack the data given by the loader,
        forward the batch through the model and compute the loss.
        Pytorch-lightning handles all the rest.

        Args:
            batch: the object returned by the loader (a list of torch.Tensor
                in most cases) but can be something else.
            _batch_nb (int): The number of the batch in the epoch.
            _train (bool): Whether in training mode. Needed only if the training
                and validation steps are fundamentally different, otherwise,
                pytorch-lightning handles the usual differences.

        Returns:
            :class:`torch.Tensor` : The loss value on this batch.
        .. note::
            This is typically the method to overwrite when subclassing
            ``System``. If the training and validation steps are somehow
            different (except for ``loss.backward()`` and ``optimzer.step()``),
            the argument ``train`` can be used to switch behavior.
            Otherwise, ``training_step`` and ``validation_step`` can be overwriten.
        """
        inputs, targets = batch
        est_targets = self(inputs)
        loss = self.loss_func(est_targets, targets)
        return loss

    def training_step(self, batch, batch_nb):
        """Pass data through the model and compute the loss.
        Backprop is **not** performed (meaning PL will do it for you).

        Args:
            batch: the object returned by the loader (a list of torch.Tensor
                in most cases) but can be something else.
            batch_nb (int): The number of the batch in the epoch.

        Returns:
            torch.Tensor, the value of the loss.
        """
        loss = self.common_step(batch, batch_nb, train=True)
        self.log("loss", loss, logger=True)
        return loss

    def validation_step(self, batch, batch_nb):
        """Need to overwrite PL validation_step to do validation.

        Args:
            batch: the object returned by the loader (a list of torch.Tensor
                in most cases) but can be something else.
            batch_nb (int): The number of the batch in the epoch.
        """
        loss = self.common_step(batch, batch_nb, train=False)
        self.log("val_loss", loss, on_epoch=True, prog_bar=True)

    def on_validation_epoch_end(self):
        """Log hp_metric to tensorboard for hparams selection."""
        hp_metric = self.trainer.callback_metrics.get("val_loss", None)
        if hp_metric is not None:
            self.trainer.logger.log_metrics(
                {"hp_metric": hp_metric}, step=self.trainer.global_step
            )

    def configure_optimizers(self):
        """Initialize optimizers, batch-wise and epoch-wise schedulers."""
        if self.scheduler is None:
            return self.optimizer

        if not isinstance(self.scheduler, (list, tuple)):
            self.scheduler = [self.scheduler]  # support multiple schedulers

        epoch_schedulers = []
        for sched in self.scheduler:
            if not isinstance(sched, dict):
                if isinstance(sched, ReduceLROnPlateau):
                    sched = {"scheduler": sched, "monitor": self.default_monitor}
                epoch_schedulers.append(sched)
            else:
                sched.setdefault("monitor", self.default_monitor)
                sched.setdefault("frequency", 1)
                # Backward compat
                if sched["interval"] == "batch":
                    sched["interval"] = "step"
                assert sched["interval"] in [
                    "epoch",
                    "step",
                ], "Scheduler interval should be either step or epoch"
                epoch_schedulers.append(sched)
        return [self.optimizer], epoch_schedulers

    def train_dataloader(self):
        """Training dataloader"""
        return self.train_loader

    def val_dataloader(self):
        """Validation dataloader"""
        return self.val_loader

    def on_save_checkpoint(self, checkpoint):
        """Overwrite if you want to save more things in the checkpoint."""
        checkpoint["training_config"] = self.config
        return checkpoint



""" Run the dummy enhancement. """
import json
import logging

import hydra
import pandas as pd
from evaluate import make_scene_listener_list
from omegaconf import DictConfig

logger = logging.getLogger(__name__)


@hydra.main(config_path=".", config_name="config")
def report_score(cfg: DictConfig) -> None:
    """Run the dummy enhancement."""

    with open(cfg.path.scenes_listeners_file, "r", encoding="utf-8") as fp:
        scenes_listeners = json.load(fp)

    results_df = pd.read_csv("scores.csv")

    # Make list of all scene listener pairs that are expected in results file
    scene_listener_pairs = make_scene_listener_list(
        scenes_listeners, cfg.evaluate.small_test
    )
    selection_df = pd.DataFrame(scene_listener_pairs, columns=["scene", "listener"])

    # Select the expected scene listener pairs from the results file
    selected_results_df = pd.merge(results_df, selection_df, on=["scene", "listener"])

    if len(selected_results_df) != len(selection_df):
        print("The following results were not found:")
        difference = pd.concat(
            [selected_results_df[["scene", "listener"]], selection_df]
        ).drop_duplicates(keep=False)
        print(difference)
    else:
        # All expected results were found so report the mean score
        print(f"Scores based on {len(selected_results_df)} scenes.")
        print(selected_results_df[["haspi", "hasqi", "combined"]].mean(axis=0))


if __name__ == "__main__":
    report_score()



"""Supplied dataclass to represent a monaural audiogram"""

from dataclasses import dataclass

import numpy as np

DEFAULT_CLARITY_CFS = [250, 500, 1000, 2000, 3000, 4000, 6000, 8000]

FULL_STANDARD_CFS = np.array(
    [
        125,
        250,
        500,
        750,
        1000,
        1500,
        2000,
        3000,
        4000,
        6000,
        8000,
        10000,
        12000,
        14000,
        16000,
    ]
)


@dataclass
class Audiogram:
    """Dataclass to represent an audiogram.

    Attributes:
        levels (ndarray): The levels for the left and right ear
        cfs (ndarray): The centre-frequencies at which the levels are measured
    """

    levels: np.ndarray
    cfs: np.ndarray = np.array(DEFAULT_CLARITY_CFS)

    @property
    def severity(self):
        """Categorise HL severity level for the audiogram.

        Note that this categorisation is different from that of the British
        Society of Audiology, which recommends descriptors mild, moderate,
        severe and profound for average hearing threshold levels at 250, 500,
        1000, 2000 and 4000 Hz of 21-40 dB HL, 41-70 dB HL, 71-95 dB HL
        and > 95 dB HL, respectively (BSA Pure-tone air-conduction and
        bone-conduction threshold audiometry with and without masking
        2018).

        Returns:
            str -- severity level, one of SEVERE, MODERATE, MILD, NOTHING

        """
        # calculate mean hearing loss between 2 & 8 kHz
        impairment_freqs = np.logical_and(2000 <= self.cfs, self.cfs <= 8000)
        tmp = self.levels[impairment_freqs]
        # Ignore any None values
        impairment_degree = np.mean(tmp[tmp is not None])

        if impairment_degree > 56:
            severity_level = "SEVERE"
        elif impairment_degree > 35:
            severity_level = "MODERATE"
        elif impairment_degree > 15:
            severity_level = "MILD"
        else:
            severity_level = "NOTHING"
        return severity_level


# Reference processing: use to check levels between original and processed,
AUDIOGRAM_REF = Audiogram(
    cfs=FULL_STANDARD_CFS,
    levels=np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
)

# mild age-related hearing loss, slightly reduced from CF, first-time aid wearers group
# (used in Stafa talk by MAS)
AUDIOGRAM_MILD = Audiogram(
    cfs=FULL_STANDARD_CFS,
    levels=np.array([5, 10, 15, 18, 19, 22, 25, 28, 31, 35, 38, 40, 40, 45, 50]),
)

# mod hearing loss based on mild N2 flat/mod sloping from Bisgaard et al. 2020
AUDIOGRAM_MODERATE = Audiogram(
    cfs=FULL_STANDARD_CFS,
    levels=np.array([15, 20, 20, 22.5, 25, 30, 35, 40, 45, 50, 55, 55, 60, 65, 65]),
)

# (mod-)severe age-related hearing loss, average of MAS/KA summer proj 2011, elderly HI
AUDIOGRAM_MODERATE_SEVERE = Audiogram(
    cfs=FULL_STANDARD_CFS,
    levels=np.array([19, 19, 28, 35, 40, 47, 52, 56, 58, 58, 63, 70, 75, 80, 80]),
)

if __name__ == "__main__":
    # test code
    audiogram = Audiogram(
        cfs=np.array([1000, 3000, 10000]), levels=np.array([10, 50, 80])
    )
    print(audiogram, audiogram.severity)
    print(Audiogram(cfs=np.array([3000]), levels=np.array([10])).severity)
    print(Audiogram(cfs=np.array([3000]), levels=np.array([20])).severity)
    print(Audiogram(cfs=np.array([3000]), levels=np.array([40])).severity)
    print(Audiogram(cfs=np.array([3000]), levels=np.array([90])).severity)
